/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/24/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_fs_opers_h__
#define __cxx_fs_opers_h__ 1

/*   std includes   */


/*   project headers   */
#include <cxx/filesystem/fs-status.h>



namespace duct {
  namespace filesystem {


    enum class copy_options {
      /*   Option group controling copy_file function effects for existing target files  */
      none = 0,
      skip_existing = 1,
      overwrite_existing = 2,
      update_existing = 4,
      
      /*   Option group controling copy function effects for sub-directories  */
      recursive = 8,
      
      /*   Option group controling copy function effects for symbolic links   */
      copy_symlinks = 16,
      skip_symlinks = 32,
      
      /*   Option group controling copy function effects for choosing the form of copying   */
      directories_only = 64,
      create_symlinks = 128,
      create_hard_links = 256
    };
    
    /*    copy_options are bitmask constants   */
    constexpr copy_options operator|(copy_options x , copy_options y) {
      return static_cast<copy_options>(static_cast<int>(x) | static_cast<int>(y));
    }
    
    constexpr copy_options operator&(copy_options x , copy_options y) {
      return static_cast<copy_options>(static_cast<int>(x) & static_cast<int>(y));
    }
    
    constexpr copy_options operator^(copy_options x , copy_options y) {
      return static_cast<copy_options>(static_cast<int>(x) ^ static_cast<int>(y)); }
    
    constexpr copy_options operator~(copy_options x) {
      return static_cast<copy_options>(~static_cast<int>(x));
    }
    
    constexpr copy_options& operator&=(copy_options& x , copy_options y) {
      x = x & y;
      return x;
    }
    
    constexpr copy_options& operator|=(copy_options& x , copy_options y) {
      x = x | y;
      return x;
    }
    
    constexpr copy_options& operator^=(copy_options & x , copy_options y) {
      x = x ^ y;
      return x;
    }

    
    /*   set permissions   */
    void permissions(const path&, perms, std::error_code&) noexcept;

    inline void permissions(const path& p, perms prms)
    {
      std::error_code ec;
      permissions(p, prms, ec);
      if(ec) throw filesystem_error("duct::filesystem::permissions(...)", p, ec);
    }
    
 
    /*  creating directory  (mkdir) */
    bool create_directory(const path&, std::error_code&) noexcept;
    bool create_directory(const path&, const path&, std::error_code&) noexcept;

    inline bool create_directory(const path& p)
    {
      std::error_code ec;
      auto res = create_directory(p, ec);
      if(ec) throw filesystem_error("duct::filesystem::create_directory(...)", p, ec);
      return res;
    }
 
    inline bool create_directory(const path& p, const path& from)
    {
      std::error_code ec;
      auto res = create_directory(p, from, ec);
      if(ec) throw filesystem_error("duct::filesystem::create_directory(...)", p, from, ec);
      return res;
    }
    
    
    /*  creating directory  (mkdir -p) */
    bool create_directories(const path&, std::error_code&) noexcept;
    
    inline bool create_directories(const path& p)
    {
      std::error_code ec;
      auto res = create_directories(p, ec);
      if(ec) throw filesystem_error("duct::filesystem::create_directories(...)", p, ec);
      return res;
    }
   

    /*   creating  links  */
    void create_symlink(const path&, const path&, std::error_code&) noexcept;
    void create_hard_link(const path&, const path&, std::error_code&) noexcept;
    void create_directory_symlink(const path&, const path&, std::error_code&) noexcept;
    
    inline void create_symlink(const path& p, const path& newlnk)
    {
      std::error_code ec;
      create_symlink(p, newlnk, ec);
      if(ec) throw filesystem_error("duct::filesystem::create_symlink(...)", p, newlnk, ec);
    }

    inline void create_hard_link(const path& p, const path& newlnk)
    {
      std::error_code ec;
      create_hard_link(p, newlnk, ec);
      if(ec) throw filesystem_error("duct::filesystem::create_hard_link(...)", p, newlnk, ec);
    }
    
    inline void create_directory_symlink(const path& p, const path& newlnk)
    {
      std::error_code ec;
      create_directory_symlink(p, newlnk, ec);
      if(ec) throw filesystem_error("duct::filesystem::create_directory_symlink(...)", p, newlnk, ec);
    }
 
    
    /*   copy symlink   */
    inline void copy_symlink(const path& existing_symlink, const path& newlnk, std::error_code& ec) noexcept
    {
      path p(read_symlink(existing_symlink, ec));
      if(ec) return;
      
      std::error_code lec;
      
      if(is_directory(existing_symlink, lec))
        create_directory_symlink(p, newlnk, ec);
      else create_symlink(p, newlnk, ec);
    }
    
    inline void copy_symlink(const path& p, const path& newlnk) {
      std::error_code ec;
      copy_symlink(p, newlnk, ec);
      if(ec) throw filesystem_error("duct::filesystem::copy_symlink(...)", p, newlnk, ec);
    }
    

    /*   working directory   */
    path current_path(std::error_code&) noexcept;
    void current_path(const path&, std::error_code&) noexcept;
    
    inline path current_path()
    {
      std::error_code ec;
      auto res = current_path(ec);
      if(ec) throw filesystem_error("duct::filesystem::current_path()", ec);
      return res;
    }
    
    inline void current_path(const path& p)
    {
      std::error_code ec;
      current_path(p, ec);
      if(ec) throw filesystem_error("duct::filesystem::current_path(...)", p, ec);
    }
    
    
    /*   remove & remove_all  (rm & rm -r)  */
    bool remove(const path&, std::error_code&) noexcept;
    std::uintmax_t remove_all(const path&, std::error_code&) noexcept;
    
    inline bool remove(const path& p)
    {
      std::error_code ec;
      auto res = remove(p, ec);
      if(ec) throw filesystem_error("duct::filesystem::remove(...)", p, ec);
      return res;
    }
    
    inline std::uintmax_t remove_all(const path& p)
    {
      std::error_code ec;
      auto res = remove_all(p, ec);
      if(ec) throw filesystem_error("duct::filesystem::remove_all(...)", p, ec);
      return res;
    }
    
   
    /*   copy a file   */
    bool copy_file(const path&, const path&, copy_options, std::error_code&) noexcept;
    
    inline bool copy_file(const path& from, const path& to, copy_options options = copy_options::none)
    {
      std::error_code ec;
      auto res = copy_file(from, to, options, ec);
      if(ec) throw filesystem_error("duct::filesystem::copy_file(...)", from, to, ec);
      return res;
    }
   
    inline bool copy_file(const path& from, const path& to, std::error_code& ec) noexcept
    {
      return copy_file(from, to, copy_options::none, ec);
    }
  
    
    /*   copy   */
    void copy(const path&, const path&, copy_options, std::error_code&) noexcept;
    
    inline void copy(const path& from, const path& to, copy_options options = copy_options::none)
    {
      std::error_code ec;
      copy(from, to, options, ec);
      if(ec) throw filesystem_error("duct::filesystem::copy(...)", from, to, ec);
    }
    
    inline void copy(const path& from, const path& to, std::error_code& ec) noexcept
    {
      copy(from, to, copy_options::none, ec);
    }
    

  }
}

#endif
