/*  Copyright (C) 2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 1/23/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __cxx_fs_path_h__
#define __cxx_fs_path_h__ 1


/*   std headers   */
#include <string>
#include <iterator>
#include <iomanip>
#include <iostream>




namespace duct
{
  namespace filesystem
  {
    class path
    {
    public:
      /*   public types   */
      using value_type = char;
      using string_type = std::string;
      
      /*   iterators    */
      class iterator;
      using const_iterator = iterator;
      
      /*   separator   */
      static constexpr value_type  preferred_separator = '/';
      
      /*  default constructors    */
      path() = default;
      
      /*   copy   */
      path(const path&) = default;
      path& operator=(const path&) = default;
      
      /*   move    */
      path(path&&) = default;
      path& operator=(path&&) = default;
      
      /*   constructors  */
      path(const value_type *);
      path(const string_type& s) : path(s.c_str()) {}
      
      template <class InputIterator>
      path(InputIterator b, InputIterator e) : path(string_type(b, e)) {}
      
      /*   assignments    */
      path& operator=(const value_type *s) { return operator=(path(s));}
      path& operator=(const string_type& s) { return operator=(path(s));}
      path& assign(const value_type *s) { return operator=(path(s));}
      path& assign(const string_type& s) { return operator=(path(s));}
      
      template <class InputIterator>
      path& assign(InputIterator b, InputIterator e) { return operator=(path(b,e));}
      
      /*   appends   */
      path& operator/=(const path& p);
      path& operator/=(const value_type *s) { return operator/=(path(s));}
      path& operator/=(const string_type& s) { return operator/=(path(s));}
      
      path& append(const value_type *s) { return this->operator/=(path(s));}
      path& append(const string_type& s) { return this->operator/=(path(s));}
      
      template <class InputIterator>
      path& append(InputIterator b, InputIterator e) { return this->operator/=(path(b, e));}
      
      /*  concatenation   */
      path& operator+=(const path& p) { _M_path += p._M_path; return *this; }
      path& operator+=(const value_type *s) { return operator+=(path(s));}
      path& operator+=(const string_type& s) { return operator+=(path(s));}
      path& operator+=(value_type c) { _M_path += c; return *this; }
      
      path& concat(const path& p) { _M_path += p._M_path; return *this; }
      path& concat(const value_type *s) { return operator+=(path(s));}
      path& concat(const string_type& s) { return operator+=(path(s));}
      path& concat(value_type c) { _M_path += c; return *this; }
      
      template <class InputIterator>
      path& concat(InputIterator begin, InputIterator end) {
        return concat(path(begin, end));
      }
      
      /*    modifiers   */
      void  clear() { _M_path.clear(); }
      void  swap(path& rhs) { _M_path.swap(rhs._M_path); }
      path& make_preferred() { return *this;}
      path& replace_extension(const path& = path());
      
      path& remove_filename() {
        auto pos = _M_path.find_last_of('/');
        if(pos != std::string::npos) _M_path.erase(_M_path.begin() + pos, _M_path.end());
        else _M_path.clear();
        return *this;
      }
      
      
      /*  query   */
      bool empty() const { return _M_path.empty(); }
      bool has_root_path() const       { return has_root_directory() || has_root_name(); }
      bool has_root_name() const       { return !root_name().empty(); }
      bool has_root_directory() const  { return !root_directory().empty(); }
      bool has_relative_path() const   { return !relative_path().empty(); }
      bool has_parent_path() const     { return !parent_path().empty(); }
      bool has_filename() const        { return !_M_path.empty(); }
      bool has_stem() const            { return !stem().empty(); }
      bool has_extension() const       { return !extension().empty(); }
      bool is_absolute() const         { return has_root_directory();}
      bool is_relative() const         { return !is_absolute(); }
      
      /*   decomposition  */
      path filename() const;
      path stem() const;
      path extension() const;
      
      /*   on UNIX system the root path, name and directory is simple,
       *   it is always '/' if it is an absolute path, otherwise ""
       */
      path root_name() const { return (_M_path[0] == '/' ? path{"/"} : path{});}
      path root_path() const { return root_name();}
      path root_directory() const { return root_name();}
      
      /*   if it is an absoulute path then it gives the relative path to the root ('/')
       *   otherwise the path itself
       */
      path relative_path() const { return (_M_path[0] == '/' ? path{_M_path.c_str()+1} : *this);}
      
      /*   parent path   */
      path parent_path() const;
      
      /*  observers   */
      const string_type&  native() const noexcept { return _M_path;}
      const value_type*   c_str() const noexcept { return _M_path.c_str(); }
      const std::string&  string() const { return _M_path; }
      
      /*  On POSIX-like systems, the generic format is the same as the native format   */
      const std::string&  generic_string() const  { return _M_path; }
      
      /*   compare (generic, lexicographical)    */
      int compare(const path& p) const noexcept { return _M_path.compare(p._M_path);}
      int compare(const std::string& s) const { return compare(path(s)); }
      int compare(const value_type *s) const  { return compare(path(s)); }
      
      /*   iterators   */
      iterator begin() const;
      iterator end() const;
      
    private:
      string_type _M_path;
    };
    
    
    /*   Non-member functions    */
    inline void swap(path& lhs, path& rhs) {
      lhs.swap(rhs);
    }
    
    inline path operator/(const path& lhs, const path& rhs) {
      return path(lhs) /= rhs;
    }

    inline path operator+(const path& lhs, const path& rhs) {
      return path(lhs) += rhs;
    }

    inline path operator+(const path& lhs, const char *rhs) {
      return path(lhs) += rhs;
    }

    inline path operator+(const path& lhs, const path::string_type& rhs) {
      return path(lhs) += rhs;
    }
    
    /*    Many comparision operators   */
    inline bool operator==(const path& lhs, const path& rhs) {
      return lhs.compare(rhs) == 0;
    }
    
    inline bool operator!=(const path& lhs, const path& rhs) {
      return lhs.compare(rhs) != 0;
    }
    
    inline bool operator< (const path& lhs, const path& rhs) {
      return lhs.compare(rhs) < 0;
    }
    
    inline bool operator<=(const path& lhs, const path& rhs) {
      return !(rhs < lhs);
    }
    
    inline bool operator> (const path& lhs, const path& rhs) {
      return rhs < lhs;
    }
    
    inline bool operator>=(const path& lhs, const path& rhs) {
      return !(lhs < rhs);
    }
    
    /*   stream operations   */
    inline std::ostream& operator<<(std::ostream& os, const path& p) {
      return os<< std::quoted(p.string());
    }
    
    inline std::istream& operator>>(std::istream& is, path& p)
    {
      std::string str;
      is >> std::quoted(str);
      p = str;
      return is;
    }
    
    
    /*   path iterator   */
    class path::iterator
    : public std::iterator<std::bidirectional_iterator_tag, const path>
    {
    public:
      /*   increment operators   */
      path::iterator& operator++();
      
      path::iterator operator++(int) {
        iterator tmp(*this);
        this->operator++();
        return tmp;
      }
      
      /*   decrement operators   */
      path::iterator& operator--();
      
      path::iterator operator--(int) {
        iterator tmp(*this);
        this->operator--();
        return tmp;
      }
      
      /*    dereference   */
      reference operator*() const { return _M_elem;}
      pointer operator->() const { return &_M_elem;}
      
      /*   EqualityComparable  */
      bool operator==(const path::iterator& rhs) const {
        return _M_path == rhs._M_path  && _M_curr == rhs._M_curr && _M_next == rhs._M_next;
      }
      
      bool operator!=(const path::iterator& rhs) const {
        return !(this->operator==(rhs));
      }
      
    private:
      /*   data members   */
      path  _M_elem;   // current element
      const path *_M_path;
      string_type::size_type _M_curr, _M_next;
      
      /*  friend declarations  */
      friend class path;
    };
    
    
    inline path::iterator path::begin() const
    {
      iterator itr;
      itr._M_path = this;
      itr._M_curr = 0;
      itr._M_next = 0;
      itr.operator++();
      return itr;
    }
    
    
    inline path::iterator path::end() const
    {
      iterator itr;
      itr._M_path = this;
      itr._M_curr = itr._M_next = _M_path.size();
      itr._M_elem = path{};
      return itr;
    }
    
    
  }  // namespace filesystem
}  // namespace duct

#endif
