/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 15/12/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __math_chebyshev_h__
#define __math_chebyshev_h__ 1



/*   std includes   */
#include <array>
#include <type_traits>
#include <cmath>
#include <algorithm>


/*   project headers   */



namespace duct {
  
  namespace __details {
    
    template<typename _Tp, bool is_ptr = std::is_pointer<_Tp>::value>
    struct cheb_check_container
    {
      static constexpr bool is_nullptr(const _Tp&) { return false;}
    };
    
    template<typename _Tp>
    struct cheb_check_container<_Tp, true>
    {
      static constexpr bool is_nullptr(const _Tp& ptr) { return nullptr == ptr;}
    };
    
  }   /*   namespace __details    */
  
  
  
  template<typename _Tp, std::size_t _Order, class _Container = const _Tp *>
  class chebyshev_series_base
  {
  public:
    /*   default constructor   */
    chebyshev_series_base() : _M_a{0.0}, _M_b{0.0}, _M_cs{} {}
    
    /*   creating from the container of the coefficients   */
    chebyshev_series_base(double a, double b, const _Container& c)
    : _M_a{a}, _M_b{b}, _M_cs{c} {}
    
    chebyshev_series_base(double a, double b, _Container&& c)
    : _M_a{a}, _M_b{b}, _M_cs{std::move(c)} {}
    
    /*   evaluate the series    */
    _Tp operator()(double) const;
    _Tp operator()(std::size_t, double) const;
    
    _Tp derivative(double) const;
    _Tp derivative_2nd(double) const;
    _Tp integral(double) const;
    
    /*   returns the boundaries of the interpolation range   */
    double xmin() const { return _M_a;}
    double xmax() const { return _M_b;}
    
  protected:
    /*   interpolation interval   */
    double _M_a, _M_b;
    
    /*   This is the container of the coefficients. It should be vctor like.
     *   It can be a pointer to an array when we use external container.
     *   The container should have at least _Order+1 element or the
     *   expressions _M_cs[0],...,_M_cs[_Order] have to make sense.
     */
    _Container _M_cs;
  };
  
  
  
  template<typename _Tp, std::size_t _Order, class _Container>
  _Tp chebyshev_series_base<_Tp, _Order, _Container>::operator()(double x) const
  {
    /*   uninitialized polynom always zero   */
    if(__details::cheb_check_container<_Container>::is_nullptr(_M_cs)) return _Tp{0.0};
    
    double y = (2.0*x - _M_a-_M_b)/(_M_b-_M_a);
    _Tp retval{0.5*_M_cs[0]};
    
    /*    trival case   */
    if(_Order == 0u) return retval;
    
    retval += y*_M_cs[1];
    if(_Order == 1U) return retval;
    
    /*    summing up the series   */
    double tmp, y2 = 2.0*y, d2 = 2.0*y*y-1.0, d1 = y;
    
    for(std::size_t i = 2; i <= _Order; ++i) {
      retval += d2*_M_cs[i];
      tmp = d2; d2 = y2*d2 - d1; d1 = tmp;
    }
    
    return retval;
  }
  
  
  template<typename _Tp, std::size_t _Order, class _Container>
  _Tp chebyshev_series_base<_Tp, _Order, _Container>::operator()(std::size_t n, double x) const
  {
    /*   uninitialized polynom always zero   */
    if(__details::cheb_check_container<_Container>::is_nullptr(_M_cs)) return _Tp{};
    
    /*   otherwise calculate the polynomial   */
    size_t eval_order = std::min(n, _Order);
    
    double y = (2.0*x - _M_a-_M_b)/(_M_b-_M_a);
    _Tp retval{0.5*_M_cs[0]};
    
    /*    trival case   */
    if(eval_order == 0u) return retval;
    
    retval += y*_M_cs[1];
    if(eval_order == 1U) return retval;
    
    /*    summing up the series   */
    double tmp, y2 = 2.0*y, d2 = 2.0*y*y-1.0, d1 = y;
    
    for(std::size_t i = 2; i <= eval_order; ++i) {
      retval += d2*_M_cs[i];
      tmp = d2; d2 = y2*d2 - d1; d1 = tmp;
    }
    
    return retval;
  }
  
  
  
  template<typename _Tp, std::size_t _Order, class _Container>
  _Tp chebyshev_series_base<_Tp, _Order, _Container>::derivative(double x) const
  {
    /*   uninitialized polynom always zero   */
    if(__details::cheb_check_container<_Container>::is_nullptr(_M_cs)) return _Tp{0.0};
    
    double y = (2.0*x - _M_a-_M_b)/(_M_b-_M_a);
    double jac = 2.0/(_M_b-_M_a);
    
    /*    trival case    */
    if(_Order == 0U) return _Tp{0.0};
    
    _Tp retval{_M_cs[1]};
    if(_Order == 1U) return jac*retval;
    
    /*    summing up the series   */
    double tmp, y2 = 2.0*y, d2 = 2.0*y, d1 = 1.0;
    
    for(std::size_t i = 2; i <= _Order; ++i) {
      retval += i*d2*_M_cs[i];
      tmp = d2; d2 = y2*d2 - d1; d1 = tmp;
    }
    
    return jac*retval;
  }
  
  
  
  template<typename _Tp, std::size_t _Order, class _Container>
  _Tp chebyshev_series_base<_Tp, _Order, _Container>::integral(double x) const
  {
    /*   uninitialized polynom always zero   */
    if(__details::cheb_check_container<_Container>::is_nullptr(_M_cs)) return _Tp{0.0};
    
    double y = (2.0*x - _M_a-_M_b)/(_M_b-_M_a), jac = 0.5*(_M_b-_M_a);
    _Tp retval{(0.5+0.5*y)*_M_cs[0]};
    
    /*   trival cases   */
    if(_Order == 0u) return jac*retval;
    
    retval += (0.5*y*y-0.5)*_M_cs[1];
    if(_Order == 1u) return jac*retval;
    
    /*    summing up the series   */
    double tmp, y2 = 2.0*y, d2 = 2.0*y*y-1.0, d1 = y, sgn = 1.0;
    
    for(std::size_t i = 2; i <= _Order; ++i) {
      retval += (y*d2/(i+1.0) - (i*d1+sgn)/(i*i-1.0))*_M_cs[i];
      tmp = d2; d2 = y2*d2 - d1; d1 = tmp; sgn *= -1;
    }
    
    return jac*retval;
  }
  
  
  template<typename _Tp, std::size_t _Order, class _Container>
  _Tp chebyshev_series_base<_Tp, _Order, _Container>::derivative_2nd(double x) const
  {
    /*   uninitialized polynom always zero   */
    if(__details::cheb_check_container<_Container>::is_nullptr(_M_cs)) return _Tp{0.0};
    
    double y = (2.0*x - _M_a-_M_b)/(_M_b-_M_a);
    double jac = 2.0/(_M_b-_M_a);
    
    /*   trival cases   */
    if(_Order < 2u) return _Tp{0.0};
    
    _Tp retval{4.0*_M_cs[2]};
    if(_Order == 2u) return (jac*jac)*retval;
    
    /*   summing up the series   */
    double tmp, y2 = 2.0*y, d2 = y*(4.0*y*y-3.0), d1 = 2.0*y*y-1.0;
    double u2 = y*(8.0*y*y-4.0), u1 = 4.0*y*y-1.0, den = y*y-1.0;
    
    if(den == 0.0) den = 1e-27;
    
    for(std::size_t i = 3; i <= _Order; ++i)
    {
      retval += _M_cs[i]*(i*((i+1)*d2-u2)/den);
      tmp = d2; d2 = y2*d2 - d1; d1 = tmp;
      tmp = u2; u2 = y2*u2 - u1; u1 = tmp;
    }
    
    return (jac*jac)*retval;
  }
  
  
  
  /*   This is simple helper class to be able to initialize the chebyshev_series from a grid.
   *   It is usefull when we know the function only in the grid points.
   */
  template<typename _Tp, std::size_t _Order>
  class chebyshev_grid
  {
  public:
    /*   simple structure for the basepoint  */
    struct point {
      point() : x{0.0}, fval{} {}
      const double x;
      _Tp fval;
    };
    
    /*   types   */
    using value_type = _Tp;
    using iterator = typename std::array<point, _Order+1>::iterator;
    using const_iterator = typename std::array<point, _Order+1>::const_iterator;
    
    
    /*   constructors   */
    chebyshev_grid() : chebyshev_grid(-1.0, 1.0) {}
    chebyshev_grid(double, double);
    template<typename _Func> chebyshev_grid(double, double, _Func&&);
    
    /*   returns the boundary of the interpolation interval  */
    double xmin() const { return _M_a;}
    double xmax() const { return _M_b;}
    
    
    /*   return the basepoint   */
    double x(std::size_t k) const { return _M_grid[k].x;}
    
    
    /*    returns the function value in the gridpoint   */
    _Tp& fval(std::size_t k) { return _M_grid[k].fval;}
    const _Tp& fval(std::size_t k) const { return _M_grid[k].fval;}
    
    
    /*   element access   */
    point& operator[](std::size_t k) { return _M_grid[k];}
    const point& operator[](std::size_t k) const { return _M_grid[k];}
    
    
    /*   iterators to be able to use range based for  */
    iterator begin() { return _M_grid.begin();}
    iterator end() { return _M_grid.end();}
    
    const_iterator begin() const { return _M_grid.begin();}
    const_iterator end() const { return _M_grid.end();}
    
    const_iterator cbegin() const { return _M_grid.begin();}
    const_iterator cend() const { return _M_grid.end();}
    
  private:
    /*   interpolation interval   */
    double _M_a, _M_b;
    
    /*   the (xk, f(xk)) pairs   */
    std::array<point, _Order+1> _M_grid;
  };
  
  
  template<typename _Tp, std::size_t _Order>
  chebyshev_grid<_Tp, _Order>::chebyshev_grid(double a, double b)
  : _M_a{a}, _M_b{b}
  {
    /*    check the interval    */
    if(_M_a >= _M_b) throw "chebyshev_grid<...> : null function interval [a,b]";
    
    double bma = 0.5*(_M_b - _M_a);
    double bpa = 0.5*(_M_b + _M_a);
    
    /*  the basepoints   */
    for(std::size_t k = 0; k <= _Order; ++k) {
      double y = std::cos(M_PI*(k+0.5)/(_Order+1.0));
      const_cast<double&>(_M_grid[k].x) = y*bma + bpa;
    }
  }
  
  
  template<typename _Tp, std::size_t _Order> template<typename _Func>
  chebyshev_grid<_Tp, _Order>::chebyshev_grid(double a, double b, _Func&& func)
  : _M_a{a}, _M_b{b}
  {
    /*    check the interval    */
    if(_M_a >= _M_b) throw "chebyshev_grid<...> : null function interval [a,b]";
    
    double bma = 0.5*(_M_b - _M_a);
    double bpa = 0.5*(_M_b + _M_a);
    
    /*  the basepoints   */
    constexpr double pif = M_PI/(_Order+1.0);
    
    for(std::size_t k = 0; k <= _Order; ++k)
    {
      const_cast<double&>(_M_grid[k].x) = bma*std::cos(pif*(k+0.5)) + bpa;
      _M_grid[k].fval = func(_M_grid[k].x);
    }
  }
  
  
  
  template<typename _Tp, std::size_t _Order>
  class chebyshev_series : public chebyshev_series_base<_Tp, _Order, std::array<double, _Order+1>>
  {
    /*   private types   */
    using _Base = chebyshev_series_base<_Tp, _Order, std::array<double, _Order+1>>;
    
  public:
    /*   types  */
    using value_type = _Tp;
    
    /*    default constructors   */
    chebyshev_series() = default;
    
    /*   from grid   */
    explicit chebyshev_series(const chebyshev_grid<_Tp, _Order>& g) { init(g);}
    
    /*   creating from a callable   */
    template<typename _Func>
    chebyshev_series(double a, double b, _Func&& func) {
      init(a, b, std::forward<_Func>(func));
    }
    
    /*   (re)initialize the polynomial  */
    void init(const chebyshev_grid<_Tp, _Order>&);
    template<typename _Func> void init(double, double, _Func&&);
  };
  
  
  template<typename _Tp, std::size_t _Order>
  void chebyshev_series<_Tp, _Order>::init(const chebyshev_grid<_Tp, _Order>& g)
  {
    _Base::_M_a = g.xmin();
    _Base::_M_b = g.xmax();
    
    /*   calculating the coefficients    */
    constexpr double fac = 2.0/(_Order + 1.0);
    constexpr double pifac = M_PI/(_Order+1.0);
    
    for(std::size_t j = 0; j <= _Order; ++j)
    {
      _Tp sum{g[0].fval*std::cos(j*0.5*pifac)};
      
      for(std::size_t k = 1; k <= _Order; ++k)
        sum += g[k].fval*std::cos(pifac*j*(k+0.5));
      
      _Base::_M_cs[j] = fac*sum;
    }
  }
  
  
  template<typename _Tp, std::size_t _Order> template<typename _Func>
  void chebyshev_series<_Tp, _Order>::init(double a, double b, _Func&& func)
  {
    _Base::_M_a = a;
    _Base::_M_b = b;
    
    /*    check the interval    */
    if(_Base::_M_a >= _Base::_M_b) throw "chebyshev_series<...> : null function interval [a,b]";
    
    double bma = 0.5*(_Base::_M_b - _Base::_M_a);
    double bpa = 0.5*(_Base::_M_b + _Base::_M_a);
    constexpr double fac = 2.0/(_Order + 1.0);
    constexpr double pifac = M_PI/(_Order+1.0);
    
    /*   function values at the basepoints   */
    _Tp f[_Order+1];
    
    for(std::size_t k = 0; k <= _Order; ++k) {
      double y = std::cos(pifac*(k+0.5));
      f[k] = func(y*bma + bpa);
    }
    
    /*   calculating the coefficients    */
    for(std::size_t j = 0; j <= _Order; ++j)
    {
      _Tp sum = 0.0;
      
      for(std::size_t k = 0; k <= _Order; ++k)
        sum += f[k]*std::cos(pifac*j*(k+0.5));
      
      _Base::_M_cs[j] = fac*sum;
    }
  }
  
  
}   /*   namespace  duct   */



#endif


