/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 10/12/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */


/*   std includes   */
#include <cmath>
#include <algorithm>

/*   project headers   */
#include "lambert.h"



namespace duct {
  
  /*     Lambert-W function
   *     (This implementation was taken from the GSL library.)
   */
  namespace __details {
    
    /* Halley iteration (eqn. 5.12, Corless et al) */
    static double lambert_w_halley_iteration(double x, double wini, unsigned int max_iters)
    {
      double w = wini;
      
      for(unsigned i = 0; i < max_iters; ++i) {
        double e = std::exp(w), p = w + 1.0;
        double t = w*e - x;
        
        /* Newton iteration */
        if(w > 0)  t = (t/p)/e;
        
        /* Halley iteration */
        else t /= e*p - 0.5*(1.0 + 1.0/p)*t;
        
        w -= t;
        
        if(std::abs(t) < 2e-15*std::max(std::abs(w), 1.0/(std::abs(p)*e)))
          return w;
      }
      
      /* should never get here */
      return w;
    }
    
    
    /* series which appears for q near zero;
     * only the argument is different for the different branches
     */
    static double lambert_w_series_eval(double r)
    {
      static const double c[12] = {
        -1.0,
        2.331643981597124203363536062168,
        -1.812187885639363490240191647568,
        1.936631114492359755363277457668,
        -2.353551201881614516821543561516,
        3.066858901050631912893148922704,
        -4.175335600258177138854984177460,
        5.858023729874774148815053846119,
        -8.401032217523977370984161688514,
        12.250753501314460424,
        -18.100697012472442755,
        27.029044799010561650
      };
      
      double t_8 = c[8] + r*(c[9] + r*(c[10] + r*c[11]));
      double t_5 = c[5] + r*(c[6] + r*(c[7]  + r*t_8));
      double t_1 = c[1] + r*(c[2] + r*(c[3]  + r*(c[4] + r*t_5)));
      
      return c[0] + r*t_1;
    }
    
  }
  
  
  double lambert_W0(double x)
  {
    /*   trivial case   */
    if(x == 0.0) return 0.0;
    
    constexpr double one_over_E = 1.0/M_E;
    double q = x + one_over_E;
    
    /* Strictly speaking it is an error for q < 0. But because of the
     * arithmetic operation connecting x and q, I am a little
     * lenient in case of some epsilon overshoot.
     */
    if(q < 1e-15) return -1.0;
    
    
    /* series near -1/E in sqrt(q) */
    if(q < 1.0e-03) return __details::lambert_w_series_eval(std::sqrt(q));
    
    
    /*   find the root numerically    */
    static constexpr unsigned MAX_ITERS = 10;
    double w;
    
    /* obtain initial approximation from series near x=0;
     * no need for extra care, since the Halley iteration
     * converges nicely on this branch
     */
    if (x < 1.0) {
      double p = std::sqrt(2.0 * M_E * q);
      w = -1.0 + p*(1.0 + p*(-1.0/3.0 + p*11.0/72.0));
    }
    
    /* obtain initial approximation from rough asymptotic */
    else {
      w = log(x);
      if(x > 3.0) w -= log(w);
    }
    
    return __details::lambert_w_halley_iteration(x, w, MAX_ITERS);
  }
  
  
  double lambert_Wm1(double x)
  {
    if(x > 0.0) return lambert_W0(x);
    else if(x == 0.0) return 0.0;
    else {
      static constexpr unsigned int MAX_ITERS = 32;
      const double one_over_E = 1.0/M_E;
      const double q = x + one_over_E;
      double w;
      
      /* Strictly speaking it is an error for q < 0. But because of the
       * arithmetic operation connecting x and q, I am a little
       * lenient in case of some epsilon overshoot.
       */
      if(q < 1e-15) return -1.0;
      
      if(x < -1.0e-6)
      {
        /* Obtain initial approximation from series about q = 0,
         * as long as we're not very close to x = 0.
         * Use full series and try to bail out if q is too small,
         * since the Halley iteration has bad convergence properties
         * in finite arithmetic for q very small, because the
         * increment alternates and p is near zero.
         */
        w = __details::lambert_w_series_eval(-std::sqrt(q));
        
        if(q < 3.0e-3) return w;
      }
      else {
        /* Obtain initial approximation from asymptotic near zero. */
        const double L_1 = std::log(-x);
        const double L_2 = std::log(-L_1);
        w = L_1 - L_2 + L_2/L_1;
      }
      
      return __details::lambert_w_halley_iteration(x, w, MAX_ITERS);
    }
  }

  
  double lambert_W0_lx(double lx)
  {
    /*   w = log(x) < 0   */
    if(lx < 0.0) return lambert_W0(std::exp(lx));
    
    /* obtain initial approximation from rough asymptotic */
    double ow, w = lx;
    if(lx > 1.1) w -= std::log(lx);
    
    do {
      ow = w;
      w = w/(1.0+w)*(1.0 + lx-std::log(w));
    } while(std::abs(w-ow) > 1e-8*w);
    
    return w;
  }

}
