/*  Copyright (C) 2002-2013 Zoltán Nagy
 *  Created by Zoltan Nagy on 3/12/02.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __duct_event_h__
#define __duct_event_h__ 1


/*   std includes   */
#include <vector>
#include <algorithm>
#include <iostream>
#include <climits>
#include <initializer_list>


/*   project headers   */
#include <bits/duct-model.h>
#include <bits/hep-lorentzvector.h>
#include <bits/hep-diracspinor.h>
#include <cxx/misc/cxx-iterator.h>



namespace duct {
  
  
  /*  This class is for storing one particle informations.
   *    => flavor        : flavor of the particle
   *    => momentum      : four mementum of the particle
   *    => plus, minus   : momentum in Dirac spinor representation
   *    => color, colorP : color connections
   */
  struct particle_state
  {
    particle_state() : flavor{0} {}
    
    /*   flavor information   */
    int flavor;
    
    /*    momentum of the particle    */
    lorentzvector<double> momentum;

    
    /*  Healthiness of a gluon:
     *    => '-' not applicable
     *    => 'f' frail gluon (or color 8 object)
     *    => 'h' healthy gluon (or color 8 object)
     */
    char health;
    
    /*   color connections   */
    struct colorlink
    {
      /*   make a dead link, it means the particle is not QCD parton   */
      static colorlink dead() {
        return {INT_MAX, INT_MIN};
      }
      
      //----- make quark a quark connection ----
      static colorlink quark(int n) {
        return {quarkend, n};
      }
      
      //----- make quark a antiquark connection ----
      static colorlink antiquark(int p) {
        return {p, antiquarkend};
      }
      
      //----- test if it is a quark -----
      bool is_quark() const { return prev == quarkend;}
      bool is_triplet() const { return prev == quarkend;}
      
      //----- test if it is a quark -----
      bool is_antiquark() const { return next == antiquarkend;}
      bool is_antitriplet() const { return next == antiquarkend;}
      
      //----- test if it is a quark -----
      bool is_gluon() const { return next > antiquarkend && prev < quarkend;}
      bool is_octet() const { return next > antiquarkend && prev < quarkend;}
      
      //----- test if it is connection ----
      bool is_qcd() const {
        return next > INT_MIN && prev < INT_MAX && !(prev == quarkend && next == antiquarkend);
      }
      
      /*   test if it is connected to parton i  */
      bool is_connected(int i) const {
        return prev == i || next == i;
      }
      
      //---- labeling the quark and antiquark ends -----
      static constexpr int quarkend =  INT_MAX-1;
      static constexpr int antiquarkend = INT_MIN+1;
      
      /*  The label of the previous and next parton on the color string.
       *  this should follow our labeling convetion (tagged integers).
       */
      int prev, next;
    } color, colorP;
  };
  
  inline std::ostream& operator<<(std::ostream& os, const particle_state::colorlink& c)
  {
    if(c.is_quark()) os<<"[Q, "<<c.next<<")";
    else if(c.is_antiquark()) os<<"("<<c.prev<<", A]";
    else if(c.is_gluon()) os<<"("<<c.prev<<","<<c.next<<")";
    else os<<"(-,-)";
    
    return os;
  }
  
  
  /*    Helper classes for implementing ranges for event   */
  namespace __details {
    
    template<std::size_t _Nin, class _Range>
    struct event_base_range;
    
    
    template<std::size_t _Nin>
    struct event_base_range<_Nin, non_qcd>
    {
      static constexpr std::size_t begin(std::size_t) { return 0u;}
      static constexpr std::size_t end(std::size_t _n_other, std::size_t) { return _n_other;}
      static constexpr std::size_t number(std::size_t _n_other, std::size_t) { return _n_other;}
    };
    
    
    template<std::size_t _Nin>
    struct event_base_range<_Nin, qcd_out>
    {
      static constexpr std::size_t begin(std::size_t _n_other) { return _n_other+_Nin;}
      static constexpr std::size_t end(std::size_t _n_other, std::size_t _n_qcd) { return _n_other+_Nin+_n_qcd;}
      static constexpr std::size_t number(std::size_t, std::size_t _n_qcd) { return _n_qcd;}
    };
    
    
    template<std::size_t _Nin>
    struct event_base_range<_Nin, incomings>
    {
      static constexpr std::size_t begin(std::size_t _n_other) { return _n_other;}
      static constexpr std::size_t end(std::size_t _n_other, std::size_t) { return _n_other+_Nin;}
      static constexpr std::size_t number(std::size_t, std::size_t) { return _Nin;}
    };
    
  
    
    /*    base class for the event classes, we used to have statistical_basis    */
    template<std::size_t _Nin>
    class event_base
    {
      /*   some constraint for the template parameter   */
      static_assert(_Nin == 1 || _Nin == 2, "Number of the incoming particles has to be either 1 or 2!!!\n");
      
      /*   number of incoming particles   */
      static constexpr std::size_t _S_number_of_in = _Nin;
      
      using _Base = std::vector<particle_state>;
      
    public:
      /*     standard public types    */
      using particle = particle_state;
      using size_type = _Base::size_type;
      using value_type = _Base::value_type;
      
      using iterator = _Base::iterator;
      using const_iterator = _Base::const_iterator;
      
      using reverse_iterator = _Base::reverse_iterator;
      using const_reverse_iterator = _Base::const_reverse_iterator;
      
      using pointer = _Base::pointer;
      using const_pointer = _Base::const_pointer;
      
      using reference = _Base::reference;
      using const_reference = _Base::const_reference;
      
      
      /*    default constructors   */
      event_base() : event_base(0u, 0u) {}
      
      /*    non standard constructor    */
      event_base(size_type _n_other, size_type _n_qcd)
      : color_suppression_index{0u}
      , _M_model{get_default_model()}
      , _M_number_of_non_qcd{_n_other}
      , _M_number_of_qcd{_n_qcd}
      {
        auto _n_part = _n_other + _S_number_of_in + _n_qcd;
        _M_particles.reserve(std::max<std::size_t>(_n_part + 10, 22));
        _M_particles.resize(_n_part);
      }
      
      
      /*   Access to the partons or parton like objects    */
      reference operator[](int k) {
        return *(_M_particles.begin() + _M_number_of_non_qcd + _S_number_of_in + k-1);
      }
      
      const_reference operator[](int k) const {
        return *(_M_particles.begin() + _M_number_of_non_qcd + _S_number_of_in + k-1);
      }
      
      
      /*     iterators     */
      iterator begin() { return _M_particles.begin();}
      iterator end() { return _M_particles.end();}
      
      const_iterator begin() const { return _M_particles.begin();}
      const_iterator end() const { return _M_particles.end();}
      
      const_iterator cbegin() const { return _M_particles.cbegin();}
      const_iterator cend() const { return _M_particles.cend();}
      
      
      /*   ranges   */
      template<class _Range> iterator_range<iterator> range() {
        return {begin() + event_base_range<_Nin, _Range>::begin(_M_number_of_non_qcd),
          begin() + event_base_range<_Nin, _Range>::end(_M_number_of_non_qcd, _M_number_of_qcd)};
      }
      
      template<class _Range> iterator_range<const_iterator> crange() const {
        return {cbegin() + event_base_range<_Nin, _Range>::begin(_M_number_of_non_qcd),
          cbegin() + event_base_range<_Nin, _Range>::end(_M_number_of_non_qcd, _M_number_of_qcd)};
      }
      
      template<class _Range> iterator_range<const_iterator> range() const {
        return crange<_Range>();
      }
      
      /*  The iteratror is not necessary a pointer, so we provide ranges with pointer   */
      template<class _Range> iterator_range<pointer> data() {
        return {_M_particles.data() + event_base_range<_Nin, _Range>::begin(_M_number_of_non_qcd),
          _M_particles.data() + event_base_range<_Nin, _Range>::end(_M_number_of_non_qcd, _M_number_of_qcd)};
      }
      
      template<class _Range> iterator_range<const_pointer> cdata() const {
        return {_M_particles.data() + event_base_range<_Nin, _Range>::begin(_M_number_of_non_qcd),
          _M_particles.data() + event_base_range<_Nin, _Range>::end(_M_number_of_non_qcd, _M_number_of_qcd)};
      }
      
      template<class _Range> iterator_range<const_pointer> data() const {
        return cdata<_Range>();
      }
      
      
      /*   number of different kind of particles    */
      template<class _Range> constexpr size_type number_of() const {
        return event_base_range<_Nin, _Range>::number(_M_number_of_non_qcd, _M_number_of_qcd);
      }
      
      /*   momentum sum over a range   */
      template<class _Range> lorentzvector<double> momentum_sum() const
      {
        lorentzvector<double> retval;
        
        for(const auto& ri : this->crange<_Range>())
          retval += ri.momentum;
        
        return retval;
      }
      
      /*    resize the array    */
      void resize(size_type _n_other, size_type _n_qcd)
      {
        _M_number_of_non_qcd = _n_other;
        _M_number_of_qcd = _n_qcd;
        _M_particles.resize(_n_other + _S_number_of_in + _n_qcd);
      }
      
      
      /*    get/set the particle table    */
      model_ref get_model() const { return _M_model;}
      void set_model(const model_ref& mdl) { _M_model = mdl;}
      
      
      /*    functions for inserting open and closed color strings   */
      void insert_open_string(const int *, const int *);
      void insert_open_stringP(const int *, const int *);
      void insert_closed_string(const int *, const int *);
      void insert_closed_stringP(const int *, const int *);
      
      /*    inserting open and closed color strings using initializer list    */
      void insert_open_string(std::initializer_list<int> ll) { insert_open_string(ll.begin(), ll.end());}
      void insert_open_stringP(std::initializer_list<int> ll) { insert_open_stringP(ll.begin(), ll.end());}
      void insert_closed_string(std::initializer_list<int> ll) { insert_closed_string(ll.begin(), ll.end());}
      void insert_closed_stringP(std::initializer_list<int> ll) { insert_closed_stringP(ll.begin(), ll.end());}
      
      
      /*   check if partons i and j are color connected in {c}_m  */
      bool are_color_connected(int i, int j) const {
        return operator[](i).color.is_connected(j);
      }
      
      /*   check if partons i and j are color connected in {c'}_m  */
      bool are_colorP_connected(int i, int j) const {
        return operator[](i).colorP.is_connected(j);
      }
      
      /*   make the health report of color 8 objects   */
      char health_report(int) const;
      void make_health_report();
      
      
    public:
      /*  This is a public data member. It represents the color supression index
       *  of the state. During the evolution we are able to keep tracing this
       *  quantity (I hope). It should be always an even number, unless we start with odd.
       */
      unsigned int color_suppression_index;

      
    private:
      /*   particle table    */
      model_ref _M_model;
      
      /*    number of the non-QCD and QCD particles    */
      size_type _M_number_of_non_qcd, _M_number_of_qcd;
      
      /*   list of particles    */
      _Base _M_particles;
    };
    
    
    /*    inserting color strings   */
    template<std::size_t _Nin>
    void event_base<_Nin>::insert_closed_string(const int *begin, const int *end)
    {
      event_base<_Nin>& p = *this;
      const int *i;
      
      p[*begin].color.prev = *(end-1);
      p[*begin].color.next = *(begin+1);
      
      for(i = begin+1; i < end-1; i++) {
        p[*i].color.prev = *(i-1);
        p[*i].color.next = *(i+1);
      }
      
      p[*i].color.prev = *(i-1);
      p[*i].color.next = *begin;
    }
    
    template<std::size_t _Nin>
    void event_base<_Nin>::insert_closed_stringP(const int *begin, const int *end)
    {
      event_base<_Nin>& p = *this;
      const int *i;
      
      p[*begin].colorP.prev = *(end-1);
      p[*begin].colorP.next = *(begin+1);
      
      for(i = begin+1; i < end-1; i++) {
        p[*i].colorP.prev = *(i-1);
        p[*i].colorP.next = *(i+1);
      }
      
      p[*i].colorP.prev = *(i-1);
      p[*i].colorP.next = *begin;
    }
    
    template<std::size_t _Nin>
    void event_base<_Nin>::insert_open_string(const int *begin, const int *end)
    {
      event_base<_Nin>& p = *this;
      const int *i;
      
      p[*begin].color.prev = particle::colorlink::quarkend;
      p[*begin].color.next = *(begin+1);
      
      for(i = begin+1; i < end-1; i++) {
        p[*i].color.prev = *(i-1);
        p[*i].color.next = *(i+1);
      }
      
      p[*i].color.prev = *(i-1);
      p[*i].color.next = particle::colorlink::antiquarkend;
    }
    
    template<std::size_t _Nin>
    void event_base<_Nin>::insert_open_stringP(const int *begin, const int *end)
    {
      event_base<_Nin>& p = *this;
      const int *i;
      
      p[*begin].colorP.prev = particle::colorlink::quarkend;
      p[*begin].colorP.next = *(begin+1);
      
      for(i = begin+1; i < end-1; i++) {
        p[*i].colorP.prev = *(i-1);
        p[*i].colorP.next = *(i+1);
      }
      
      p[*i].colorP.prev = *(i-1);
      p[*i].colorP.next = particle::colorlink::antiquarkend;
    }
    
    
    template<std::size_t _Nin>
    inline char event_base<_Nin>::health_report(int l) const
    {
      //---- the health is defined only for color 8 objects -----
      event_base<_Nin>& p = *this;
      
      if(!(p[l].color.is_octet())) return '-';
      
      //---- it is a color 8, find the heath status -----
      int rprev = l, lnext;
      do {
        if((lnext = p[rprev].color.next) == l) return 'f';
        if((rprev = p[lnext].colorP.prev) == l) return 'h';
      } while(true);
    }
    
    
    template<std::size_t _Nin>
    inline void event_base<_Nin>::make_health_report()
    {
      event_base<_Nin>& p = *this;
      
      for(int l = 1; l <= (int)_M_number_of_qcd; ++l)
      {
        //---- the health is defined only for color 8 objects -----
        if(!(p[l].color.is_octet())) p[l].health = '-';
        
        //---- it is a color 8, find the heath status -----
        else {
          int rprev = l, lnext;
          do {
            if((lnext = p[rprev].color.next) == l) { p[l].health = 'f'; break;}
            if((rprev = p[lnext].colorP.prev) == l) { p[l].health = 'h'; break;}
          } while(true);
        }
      }
    }
    
    
    template<std::size_t _Nin>
    std::ostream& operator<<(std::ostream& os, const event_base<_Nin>& a)
    {
      model_ref mdl = a.get_model();
      
      //----- the ranges -----
      auto qcd = a.template range<qcd_out>();
      auto non = a.template range<non_qcd>();
      auto inc = a.template range<incomings>();
      
      //------ print the momenta ----
      int idx = 1-a.template number_of<non_qcd>() - a.template number_of<incomings>();
      
      if(non.begin() < non.end()) {
        os<<"+---- non-qcd outgoings -----\n";
        for(auto& pi : non)
          os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "
          <<pi.momentum<<"  "<<pi.momentum.mag2()<<'\n';
      }
      
      os<<"+---- incomings -----\n";
      for(auto& pi : inc)
        os<<"|  "<<mdl.name(-(pi.flavor))<<"("<<idx++<<") : "
        <<pi.momentum<<"  "<<pi.momentum.mag2()
        <<"\t   "<<pi.health
        <<"  "<<pi.color
        <<"  "<<pi.colorP
        <<'\n';
      
      os<<"+---- qcd outgoings -----\n";
      for(auto& pi : qcd)
        os<<"|  "<<mdl.name(pi.flavor)<<"("<<idx++<<") : "
        <<pi.momentum<<"  "<<pi.momentum.mag2()
        <<"\t   "<<pi.health
        <<"  "<<pi.color
        <<"  "<<pi.colorP
        <<'\n';
      
      os<<"+----\n\n\n";
      
      auto itot = a.template momentum_sum<incomings>();
      auto otot = a.template momentum_sum<non_qcd>()+a.template momentum_sum<qcd_out>();
      
      os<<"Total incoming momentum     : "<<itot<<'\n';
      os<<"Total outgoing momentum     : "<<otot<<'\n';
      os<<"Momentum conservation       : "<<itot-otot<<"\n\n";
      os<<"Color supression index      : "<<a.color_suppression_index<<std::endl;
      
      return os;
    }
    
  }   //   namespace __details
  
  
  
  
  /* This class helps to decribe the incoming object.
   * We need more information if the incoming particle 
   * is composite object like a hadron.
   */
  struct colliding_hadron {
    /* The lorentz momentum of the incoming particle*/
    lorentzvector<double> momentum;
    
    /* Type of the hadron (PDG key) */
    int key;
  };


  /* Empty template declaration of the class event.
   * This class is inhereted from the statistical_basis class
   * and add extra, process dependent properties to it.
   */
  template<class _Inh>
  class event; 
  
 
  /*  Definition of the event class for the decay processes. 
   *  In the decay processes we have only one incoming particle.
   *  Here we consider only elementary particle decay.
   */
  template<>
  class event<decay> : public __details::event_base<1u>
  {
    using _Base = __details::event_base<1u>;
    
  public:
    /*    default constructors   */
    event() = default;
    
    /*    non standard constructor    */
    event(size_type _n_other, size_type _n_qcd)
    : _Base{_n_other, _n_qcd} {}

    /*    merge back the decay subevent   */
    event& merge(int, const event&);
  };
  
  
  /*  Definition of the event class for e+e- like processes 
   *  (the colliding particles are elementary). We have two incomings.
   */
  template<>
  class event<epa> : public __details::event_base<2u>
  {
    using _Base = __details::event_base<2u>;
    
  public:
    /*    default constructors   */
    event() = default;
    
    /*    non standard constructor    */
    event(size_type _n_other, size_type _n_qcd)
    : _Base{_n_other, _n_qcd} {}
    
    /*    merge back the decay subevent   */
    event& merge(int, const event&);
  };
  
  
  /*  Definition of the event class for DIS like processes, one elementary (electron) 
   *  and one hadron in the initial state. Here we have to provide pdf for the hadron.
   */
  template<>
  class event<dis> : public __details::event_base<2u>
  {
    using _Base = __details::event_base<2u>;
    
  public:
    /*    default constructors   */
    event() = default;
    
    /*    non standard constructor    */
    event(size_type _n_other, size_type _n_qcd)
    : _Base{_n_other, _n_qcd} {}
    
    /*    acces to the hadron objects     */
    colliding_hadron& operator[](const hadron& i) {return _M_hadron;}
    const colliding_hadron& operator[](const hadron& i) const { return _M_hadron;}

    /*   access to the partons   */
    using event_base<2u>::operator[];
    
    /*    merge back the decay subevent   */
    event& merge(int, const event&);
    
  private:
    /*    colliding hadron    */
    colliding_hadron _M_hadron;
  };

  
  /*  Definitions of the event class for hadron-hadron collision. We have two hadron 
   *  in the initial state and the corresponding pdf should be provided. 
   */
  template<>
  class event<hhc> : public __details::event_base<2u>
  {
    using _Base = __details::event_base<2u>;
    
  public:
    /*    default constructors   */
    event() = default;
    
    /*    non standard constructor    */
    event(size_type _n_other, size_type _n_qcd)
    : _Base{_n_other, _n_qcd} {}
    
    /*    acces to the hadron objects     */
    colliding_hadron& operator[](const hadron& i) {return _M_hadron[i.idx+1];}
    const colliding_hadron& operator[](const hadron& i) const { return _M_hadron[i.idx+1];}
    
    /*   access to the partons   */
    using event_base<2u>::operator[];
    
    /*    merge back the decay subevent   */
    event& merge(int, const event&);
    
  private:
    /*    colliding hadron    */
    colliding_hadron _M_hadron[2];
  };
  
  
  
  //  FIXME:  this is very simple output operator
  template<class _Tag> std::ostream& operator<<(std::ostream&, const event<_Tag>&);
  
  
  template<> inline
  std::ostream& operator<<(std::ostream& os, const event<decay>& a) {
    return os<<static_cast<const __details::event_base<1u>&>(a);
  }
  
  template<> inline
  std::ostream& operator<<(std::ostream& os, const event<epa>& a) {
    return os<<static_cast<const __details::event_base<2u>&>(a);
  }
  
  template<> inline
  std::ostream& operator<<(std::ostream& os, const event<dis>& a) 
  {
    os<<"H : "<<a[hadron(0)].momentum<<"  "<<a[hadron(0)].momentum.mag2()<<std::endl;
    return os<<static_cast<const __details::event_base<2u>&>(a);
  }
 
  template<> inline
  std::ostream& operator<<(std::ostream& os, const event<hhc>& a) 
  {
    os<<"A : "<<a[hadron(-1)].momentum<<"  "<<a[hadron(0)].momentum.mag2()<<std::endl;
    os<<"B : "<<a[hadron(0)].momentum<<"  "<<a[hadron(0)].momentum.mag2()<<std::endl;
    return os<<static_cast<const __details::event_base<2u>&>(a);
  }
}

#endif
