/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_hep_dist__
#define __deductor_hep_dist__ 1


//   Standard includes
#include <list>
#include <vector>
#include <iostream>

//   deductor includes
#include <bits/hep-sample_statistic.h>
#include <cxx/binary-io.h>


namespace duct {
  
 
  /*   Some predefined goodies to implement 
   *   simple distributions and histograms.
   */
  namespace distpoint {
    
    template<class _Point>
    struct traits
    {
      typedef _Point point_type;
      
      /*   fancy printing   */
      static void print(std::ostream& os,  const point_type& p) { os<<p;}
    };
    
    template<>
    struct traits<void>
    {
      typedef void point_type;
    };
    
    template<> inline
    void traits<float>::print(std::ostream& os,  const float& p) {
      if(p <= 0.0f) os<<" ";
      os<<p;
    }
    
    template<> inline
    void traits<double>::print(std::ostream& os,  const double& p) {
      if(p <= 0.0) os<<" ";
      os<<p;
    }
    
    template<> inline
    void traits<long double>::print(std::ostream& os,  const long double& p) {
      if(p <= 0.0l) os<<" ";
      os<<p;
    }
    
    template<> inline
    void traits<int>::print(std::ostream& os,  const int& p) {
      if(p <= 0) os<<" ";
      os<<p;
    }
    
    template<> inline
    void traits<long int>::print(std::ostream& os,  const long int& p) {
      if(p <= 0) os<<" ";
      os<<p;
    }
    
    template<> inline
    void traits<long long int>::print(std::ostream& os,  const long long int& p) {
      if(p <= 0) os<<" ";
      os<<p;
    }
    
  
    //    a simple point type for the 1D histograms
    struct hist1d {
      double xmin, xmax;
    };
    
    inline std::ostream& write(std::ostream& os, const hist1d& p) {
      return duct::write(os, p.xmin, p.xmax);
    }
    
    inline std::istream& read(std::istream& is, hist1d& p) {
      return duct::read(is, p.xmin, p.xmax);
    }

    
    template<>
    struct traits<hist1d>
    {
      typedef hist1d point_type;

      /*   fancy printing   */
      static void print(std::ostream& os,  const point_type& p) {
        if(p.xmin >= 0) os<<" ";
        os<<p.xmin<<"  ";
        
        if(0.5*(p.xmin+p.xmax) >= 0) os<<" ";
        os<<0.5*(p.xmin+p.xmax)<<"  ";
        
        if(p.xmax >= 0) os<<" ";
        os<<p.xmax<<" ";
      }
    };
    
    
    //    a simple point type for the 2D histograms
    struct hist2d {
      double xmin, xmax;
      double ymin, ymax;
    };
    
    inline std::ostream& write(std::ostream& os, const hist2d& p) {
      return duct::write(os, p.xmin, p.xmax, p.ymin, p.ymax);
    }
    
    inline std::istream& read(std::istream& is, hist2d& p) {
      return duct::read(is, p.xmin, p.xmax, p.ymin, p.ymax);
    }
    
    
    template<>
    struct traits<hist2d>
    {
      typedef hist2d point_type;
      
      /*   fancy printing   */
      static void print(std::ostream& os,  const point_type& p) {
        if(p.xmin >= 0) os<<" ";
        os<<p.xmin<<"  ";
        
        if(0.5*(p.xmin+p.xmax) >= 0) os<<" ";
        os<<0.5*(p.xmin+p.xmax)<<"  ";
        
        if(p.xmax >= 0) os<<" ";
        os<<p.xmax<<"  ";
        
        if(p.ymin >= 0) os<<" ";
        os<<p.ymin<<"  ";
        
        if(0.5*(p.ymin+p.ymax) >= 0) os<<" ";
        os<<0.5*(p.ymin+p.ymax)<<"  ";
        
        if(p.ymax >= 0) os<<" ";
        os<<p.ymax<<" ";
      }
    };
  
    
    /*   step function  */
    struct step
    {
      double operator()(double y, double x) const  {
        return x > y ? 1.0 : 0.0;
      }

      double operator()(const hist1d& y, double x) const {
        return x > y.xmin ? 1.0 : 0.0;
      }
      
      double operator()(const hist2d& h, double x, double y) const
      {
        if(x <= h.xmin) return 0.0;
        if(y <= h.ymin) return 0.0;
        return 1.0;
      }
    };
    
    
    /*   simple apprixmation of the Dirac delta function  */
    struct dirac
    {
      double operator()(const hist1d& y, double x) const {
        return y.xmin <= x && x < y.xmax ? 1.0/(y.xmax-y.xmin) : 0.0;
      }
      
      double operator()(const hist2d& h, double x, double y) const
      {
        if(!(h.xmin <= x && x < h.xmax)) return 0.0;
        if(!(h.ymin <= y && y < h.ymax)) return 0.0;
        return 1.0/(h.xmax-h.xmin)/(h.ymax-h.ymin);
      }
    };
    
    struct stepdirac
    {
      double operator()(const hist2d& h, double x, double y) const
      {
        if(x <= h.xmin) return 0.0;
        if(!(h.ymin <= y && y < h.ymax)) return 0.0;
        return 1.0/(h.ymax-h.ymin);
      }
    };
    
    struct diracstep
    {
      double operator()(const hist2d& h, double x, double y) const
      {
        if(!(h.xmin <= x && x < h.xmax)) return 0.0;
        if(y <= h.ymin) return 0.0;
        return 1.0/(h.xmax-h.xmin);
      }
    };
    
    
    /*   simple class to setup distpoints   */
    template<typename _Tp>
    struct spacing;
    
    
    template<>
    struct spacing<float> {
      /*   linear spacing   */
      static std::vector<float> linear(unsigned, float, float);
      
      /*   logarithmic spacing   */
      static std::vector<float> logarithmic(unsigned, float, float);
    };
    
  
    template<>
    struct spacing<double> {
      /*   linear spacing   */
      static std::vector<double> linear(unsigned, double, double) ;
      
      /*   logarithmic spacing   */
      static std::vector<double> logarithmic(unsigned, double, double);
    };
    
    
    template<>
    struct spacing<long double> {
      /*   linear spacing   */
      static std::vector<long double> linear(unsigned, long double, long double);
      
      /*   logarithmic spacing   */
      static std::vector<long double> logarithmic(unsigned, long double, long double);
    };
    
    
    template<>
    struct spacing<hist1d> {
      /*   linear spacing   */
      static std::vector<hist1d> linear(unsigned, double, double);
      
      /*   logarithmic spacing   */
      static std::vector<hist1d> logarithmic(unsigned, double, double);

      /*   manual spacing   */
      static std::vector<hist1d> manual(std::initializer_list<double>);
    };
    
    
    template<>
    struct spacing<hist2d> {
      /*   linear spacing   */
      static std::vector<hist2d>
      linear(unsigned, double, double, unsigned, double, double);
      
      /*   logarithmic spacing   */
      static std::vector<hist2d>
      logarithmic(unsigned, double, double, unsigned, double, double);
      
      /*   logarithmic-linear spacing   */
      static std::vector<hist2d>
      loglinear(unsigned, double, double, unsigned, double, double);
      
      /*   linear-logarithmic spacing   */
      static std::vector<hist2d>
      linearlog(unsigned, double, double, unsigned, double, double);
    };
  }   //  namespace distpoint
}   //   namespace duct




namespace duct {
  
  //   Forward declaration of the class distriburion
  template<typename _Tp, typename _Point = void,
           class _Traits = sample_traits<_Tp>,
           class _PointTraits = distpoint::traits<_Point> >
  class distribution;


  
  //
  //   Specialization : class _Point = void
  //
  template<typename _Tp, class _Traits>
  class distribution<_Tp, void, _Traits, distpoint::traits<void> >
  {
  public:
    //       types
    typedef _Tp sample_type;
    
    //   constructors
    distribution() { sample_statistic<_Tp>::zero(_M_tmp);}
    
    /*   computed assinment   */
    distribution& operator+=(const distribution& x) {
      _M_acc += x._M_acc;
      return *this;
    }

    /*   collecting weights   */
    void collect(const sample_type& weight) {
      sample_statistic<_Tp>::assadd(_M_tmp, weight);
    }
    
    /*     accumulate weights   */
    void accumulate() {
      _M_acc.accumulate(_M_tmp);
      reset_tmp();
    }
    
    void accumulate(const sample_type& weight) {
      sample_statistic<_Tp>::assadd(_M_tmp, weight);
      _M_acc.accumulate(_M_tmp);
      reset_tmp();
    }

    /*     reset   */
    void reset_tmp() { sample_statistic<_Tp>::zero(_M_tmp);}
    void reset() { _M_acc.reset(); reset_tmp();}
    
    /*  access to the results   */
    sample_statistic<_Tp>& get_accumulator() { return _M_acc;}
    const sample_statistic<_Tp>& get_accumulator() const { return _M_acc;}
    
  private:
    sample_type _M_tmp;
    sample_statistic<_Tp> _M_acc;
  };
  
  //   I/O functions for _Point = void
  template<typename _Tp, class _Traits> inline std::ostream&
  write(std::ostream& os, const distribution<_Tp, void, _Traits, distpoint::traits<void>>& x) {
    return write(os, x.get_accumulator());
  }

  template<typename _Tp, class _Traits> inline std::istream&
  read(std::istream& is, distribution<_Tp, void, _Traits, distpoint::traits<void>>& x) {
    read(is, x.get_accumulator());
    x.reset_tmp();
    return is;
  }
  
  template<typename _Tp, class _Traits> inline std::ostream& 
  print(std::ostream& os, const distribution<_Tp, void, _Traits, distpoint::traits<void>>& x) {
    _Traits::print(os, x.get_accumulator());
    os<<"\n";
    return os;
  }


  //
  //    General template class 
  //
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  class distribution
  {
  private:
    //    private types
    typedef distribution<_Tp, void, _Traits, distpoint::traits<void>> _Dist;
    typedef std::pair<_Point, _Dist> _Value;
    typedef std::vector<_Value> _Base;

  public:
    //   public types
    typedef _Tp sample_type;
    typedef _Point point_type;
    typedef _PointTraits point_traits;

    typedef typename _Base::iterator iterator;
    typedef typename _Base::reference reference;
    typedef typename _Base::const_iterator const_iterator;
    typedef typename _Base::const_reference const_reference;
  
    typedef typename _Base::reverse_iterator reverse_iterator;
    typedef typename _Base::const_reverse_iterator const_reverse_iterator;
 
    /*   constructors, destructor   */
    distribution() : _M_isnorm(false) {}
    
    /*   constructors   */
    template<class _InputIterator>
    distribution(_InputIterator first, _InputIterator last, bool norm = false)
    : _M_isnorm(norm) {
      for(auto ib = first; ib != last; ++ib)
        _M_base.push_back(std::make_pair(*ib, _Dist{}));
    }
    
    distribution(unsigned int n, const _Point *base, bool norm = false)
    : distribution(base, base+n, norm) {}
        
    distribution(std::initializer_list<_Point> base, bool norm = false)
    : distribution(base.begin(), base.end(), norm) {}
    
    template<class _Container>
    explicit distribution(const _Container& base, bool norm = false)
    : distribution(base.begin(), base.end(), norm) {}
    
    
    /*   computed assignment   */
    distribution& operator+=(const distribution& x) {
      const_iterator ix = x._M_base.begin();
      for(iterator i = _M_base.begin(); i < _M_base.end(); i++, ix++)
        i->second += ix -> second;
      
      if(_M_isnorm && x._M_isnorm) _M_norm += x._M_norm;
      else if(x._M_isnorm) _M_norm = x._M_norm;
      
      _M_isnorm |= x._M_isnorm;
      
      return *this;
    }
   
    /*   collecting the weights    */
    template<class _Func>
    void collect(_Func func) {
      for(auto& is : _M_base)
        is.second.collect(func(is.first));
    }
    
    template<class _Func, typename ..._Arg>
    void collect(const _Tp& w, _Func&& func, _Arg&&... x) {
      for(auto& is : _M_base) {
        auto resf = func(is.first, std::forward<_Arg>(x)...);
        if(resf != 0.0) is.second.collect(sample_statistic<_Tp>::mul(w, resf));
      }
    }
    
    /*   accumulate the collected weights   */
    void accumulate() {
      for(auto& is : _M_base) is.second.accumulate();
    }
    
    template<class _Func>
    void accumulate(_Func&& func) {
      this->collect(std::forward<_Func>(func));
      this->accumulate();
    }
     
    template<class _Func, typename ..._Arg>
    void accumulate(const _Tp& w, _Func&& func, _Arg&&... x) {
      this->collect(w, std::forward<_Func>(func), std::forward<_Arg>(x)...);
      this->accumulate();
    }
    
    
    /*   collecting and accumulating weights for the normalization   */
    void collect_norm(const sample_type& weight) { _M_norm.collect(weight);}
    void accumulate_norm(const sample_type& weight) { _M_norm.accumulate(weight);}
    void accumulate_norm() { if(_M_isnorm) _M_norm.accumulate();}

    
    /*   reinitialize the distribution   */
    void reset() { for(auto& is : _M_base) is.second.reset(); _M_norm.reset();}
    void reset_tmp() { for(auto& is : _M_base) is.second.reset_tmp(); _M_norm.reset_tmp();}

    //  capacity
    unsigned int size() const { return _M_base.size();}
    void resize(unsigned int n) { _M_base.resize(n);}
    
    /*  iterators   */
    iterator begin() { return _M_base.begin();}
    iterator end() { return _M_base.end();}
    
    const_iterator begin() const { return _M_base.begin();}
    const_iterator end() const { return _M_base.end();}

    const_iterator cbegin() const { return _M_base.cbegin();}
    const_iterator cend() const { return _M_base.cend();}

    /*   reverse iterator   */
    reverse_iterator rbegin() { return _M_base.rbegin();}
    reverse_iterator rend() { return _M_base.rend();}
    
    const_reverse_iterator rbegin() const { return _M_base.rbegin();}
    const_reverse_iterator rend() const { return _M_base.rend();}
    
    const_reverse_iterator crbegin() const { return _M_base.crbegin();}
    const_reverse_iterator crend() const { return _M_base.crend();}
    
    /*   element access   */
    reference operator[](unsigned int i) { return _M_base[i];}
    const_reference operator[](unsigned int i) const { return _M_base[i];}
    
    /*   access to the normalization   */
    _Dist& norm() { return _M_norm;}
    const _Dist& norm() const { return _M_norm;}
    
    /*   check if it is normalized distribution   */
    bool is_normalized() const { return _M_isnorm;}
    
    /*   make it normalized distribution or reset normalization   */
    void normalize(const _Dist& nn = _Dist{}) { _M_norm = nn; _M_isnorm = true;}
   
    /*    I/O  operations   */
    void write(std::ostream& os) const {
      duct::write(os, _M_base, _M_isnorm, _M_norm);
    }

    void read(std::istream& is) {
      duct::read(is, _M_base, _M_isnorm, _M_norm);
      reset_tmp();
    }
   
  private:
    /*   data members   */
    _Base _M_base;
    
    /*    normalization  */
    bool _M_isnorm;
    _Dist _M_norm;
  };

    
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  inline distribution<_Tp, _Point, _Traits, _PointTraits>
  operator+(const distribution<_Tp, _Point, _Traits, _PointTraits>& x, const distribution<_Tp, _Point, _Traits, _PointTraits>& y)
  {
    return distribution<_Tp, _Point, _Traits, _PointTraits>(x) += y;
  }
  
    
  //      I/O operations
  template<typename _Tp, class _Point, class _Traits, class _PointTraits> inline
  std::ostream& write(std::ostream& os, const distribution<_Tp, _Point, _Traits, _PointTraits>& d) {
    d.write(os);
    return os;
  }
  
  template<typename _Tp, class _Point, class _Traits, class _PointTraits> inline
  std::istream& read(std::istream& is, distribution<_Tp, _Point, _Traits, _PointTraits>& d) {
    d.read(is);
    return is;
  }

  template<class _Tp, typename _Point, class _Traits, class _PointTraits> 
  std::ostream& print(std::ostream& os, const distribution<_Tp, _Point, _Traits, _PointTraits>& d)
  {
    const auto& norm = d.norm().get_accumulator();
    
    for(auto i = d.begin(); i < d.end(); i++) {
      _PointTraits::print(os, i->first); os<<"  ";
      
      if(d.is_normalized()) _Traits::print(os, i->second.get_accumulator(), norm);
      else _Traits::print(os, i->second.get_accumulator());
      
      os<<"\n";
    }
    
    return os;
  }
  
}   //   namespace duct


#endif
