/*  Copyright (C) 2002-2013 Zoltán Nagy
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __deductor_hep_distbook_h__
#define __deductor_hep_distbook_h__ 1


//   Standard includes
#include <map>
#include <string>
#include <tuple>
#include <fstream>

//   deductor includes
#include <bits/hep-dist.h>


namespace duct {
 
  
  template<typename _Tp, class _Point = void,
    class _Traits = sample_traits<_Tp>,
    class _PointTraits = distpoint::traits<_Point>>
  class distbook
  {
  public:
    /*   public typedefs   */
    typedef _Tp sample_type;
    typedef _Point point_type;
    typedef _Traits sample_traits;
    typedef _PointTraits point_traits;
    typedef distribution<_Tp, _Point, _Traits, _PointTraits> distribution_type;
    typedef std::tuple<std::string, distribution_type> entry_type;

    typedef typename std::map<int, entry_type>::value_type        value_type;
    typedef typename std::map<int, entry_type>::reference         reference;
    typedef typename std::map<int, entry_type>::const_reference   const_reference;
    typedef typename std::map<int, entry_type>::iterator          iterator;
    typedef typename std::map<int, entry_type>::const_iterator    const_iterator;

    /*  computed assignment   */
    distbook& operator+=(const distbook& x) {
      auto iter = _M_book.begin();
      auto xiter = x.begin();
      while(iter != _M_book.end())
        std::get<1>((iter++)->second) += std::get<1>((xiter++)->second);
      return *this;
    }

    /*    create an entry in the book   */
    void create(int key, const char *title, unsigned n, const _Point *base, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(n, base, norm));
    }
 
    template<class _InputIterator>
    void create(int key, const char *title, _InputIterator first, _InputIterator last, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(first, last, norm));
    }
    
    void create(int key, const char *title, std::initializer_list<_Point> base, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(base, norm));
    }

    template<class _Container>
    void create(int key, const char *title, const _Container& base, bool norm = false) {
      _M_book[key] = entry_type(title, distribution_type(base, norm));
    }

    /*   collecting the weights    */
    template<class _Func>
    void collect(int key, _Func&& func) {
      std::get<1>(_M_book.find(key)->second).collect(std::forward<_Func>(func));
    }
    
    template<class _Func, typename ..._Arg>
    void collect(int key, const _Tp& w, _Func&& func, _Arg&&... x) {
      std::get<1>(_M_book.find(key)->second).collect(w, std::forward<_Func>(func), std::forward<_Arg>(x)...);
    }
    
    /*   accumulate the collected weights   */
    void accumulate(int key) {
      std::get<1>(_M_book.find(key)->second).accumulate();
    }
    
    template<class _Func>
    void accumulate(int key, _Func&& func) {
      std::get<1>(_M_book.find(key)->second).accumulate(std::forward<_Func>(func));
    }
    
    template<class _Func, typename ..._Arg>
    void accumulate(int key, const _Tp& w, _Func&& func, _Arg&&... x) {
      std::get<1>(_M_book.find(key)->second).accumulate(w, std::forward<_Func>(func), std::forward<_Arg>(x)...);
    }
    
    
    /*   collecting and accumulating weights for the normalization   */
    void collect_norm(int key, const _Tp& w) { std::get<1>(_M_book.find(key)->second).collect_norm(w);}
    void accumulate_norm(int key, const _Tp& w) { std::get<1>(_M_book.find(key)->second).accumulate_norm(w);}
    void accumulate_norm(int key) { std::get<1>(_M_book.find(key)->second).accumulate_norm();}
    
    
    /*   accumulate the weights in the whole book   */
    void accumulate() {
      for(auto& d : _M_book) {
        std::get<1>(d.second).accumulate();
        std::get<1>(d.second).accumulate_norm();
      }
    }
    
    /*  element access (this can also create an entry) */
    entry_type& operator[](int key) { return _M_book[key];}
    
    /*    iterators  */
    iterator begin() { return _M_book.begin();}
    iterator end() { return _M_book.end();}
    
    const_iterator begin() const { return _M_book.begin();}
    const_iterator end() const { return _M_book.end();}
    
    const_iterator cbegin() const { return _M_book.cbegin();}
    const_iterator cend() const { return _M_book.cend();}

    /*   structural information   */
    unsigned int size() const { return _M_book.size();}
    
    /*   clear the whole book   */
    void clear() { _M_book.clear();}
  
    /*    I/O operations    */
    void read(std::istream& is) { duct::read(is, _M_book);}
    void write(std::ostream& os) const { duct::write(os, _M_book);}
    
  private:
    /*    This is a map of distributions   */
    std::map<int, entry_type> _M_book;
  };
  
  
  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  inline distbook<_Tp, _Point, _Traits, _PointTraits>
  operator+(const distbook<_Tp, _Point, _Traits, _PointTraits>& x, const distbook<_Tp, _Point, _Traits, _PointTraits>& y) {
    return distbook<_Tp, _Point, _Traits, _PointTraits>(x) += y;
  }
  
  //   I/O operations
  template<typename _Tp, class _Point, class _Traits, class _PointTraits> inline
  std::ostream& write(std::ostream& os, const distbook<_Tp, _Point, _Traits, _PointTraits>& x) {
    x.write(os);
    return os;
  }
    
  template<typename _Tp, class _Point, class _Traits, class _PointTraits> inline
  std::istream& read(std::istream& is, distbook<_Tp, _Point, _Traits, _PointTraits>& x) {
    x.read(is);
    return is;

  }
  

  template<typename _Tp, class _Point, class _Traits, class _PointTraits>
  std::ostream& print(std::ostream& os, const distbook<_Tp, _Point, _Traits, _PointTraits>& x)
  {
    for(auto& vi : x) {
      os<<"#  "<<std::get<0>(vi.second)<<"\n";
      print(os, std::get<1>(vi.second))<<"\n\n";
    }
    
    return os;
  }
 
  
  
  template<typename _Tp, class _Traits>
  class distbook<_Tp, void, _Traits, distpoint::traits<void>>
  {
  public:
    /*   public typedefs   */
    typedef _Tp sample_type;
    typedef void point_type;
    typedef _Traits sample_traits;
    typedef distpoint::traits<void> point_traits;
    typedef distribution<_Tp, void, _Traits, distpoint::traits<void>> distribution_type;
    typedef std::tuple<std::string, distribution_type> entry_type;
    
    typedef typename std::map<int, entry_type>::value_type        value_type;
    typedef typename std::map<int, entry_type>::reference         reference;
    typedef typename std::map<int, entry_type>::const_reference   const_reference;
    typedef typename std::map<int, entry_type>::iterator          iterator;
    typedef typename std::map<int, entry_type>::const_iterator    const_iterator;
    
    /*  computed assignment   */
    distbook& operator+=(const distbook& x) {
      auto iter = _M_book.begin();
      auto xiter = x.begin();
      while(iter != _M_book.end())
        std::get<1>((iter++)->second) += std::get<1>((xiter++)->second);
      return *this;
    }
    
    /*    create an entry in the book   */
    void create(int key, const char *title) {
      _M_book[key] = entry_type(title, distribution_type());
    }
    
    /*   collecting the weights    */
    void collect(int key, const _Tp& w) {
      std::get<1>(_M_book.find(key)->second).collect(w);
    }
    
    /*   accumulate the collected weights   */
    void accumulate(int key) {
      std::get<1>(_M_book.find(key)->second).accumulate();
    }
    
    void accumulate(int key, const _Tp& w) {
      std::get<1>(_M_book.find(key)->second).accumulate(w);
    }
    
    /*   accumulate the weights in the whole book   */
    void accumulate() { for(auto& d : _M_book) std::get<1>(d.second).accumulate();}
    
    /*  element access (this can also create an entry) */
    entry_type& operator[](int key) { return _M_book[key];}
    
    /*    iterators  */
    iterator begin() { return _M_book.begin();}
    iterator end() { return _M_book.end();}
    
    const_iterator begin() const { return _M_book.begin();}
    const_iterator end() const { return _M_book.end();}
    
    const_iterator cbegin() const { return _M_book.cbegin();}
    const_iterator cend() const { return _M_book.cend();}
    
    /*   structural information   */
    unsigned int size() const { return _M_book.size();}
    
    /*   clear the whole book   */
    void clear() { _M_book.clear();}

    /*    I/O operations    */
    void read(std::istream& is) { duct::read(is, _M_book);}
    void write(std::ostream& os) const { duct::write(os, _M_book);}
    
  private:
    /*    This is a map of distributions   */
    std::map<int, entry_type> _M_book;
  };
  

  /*   Explicit instatiation   */
  extern template class distbook<double, void>;
  extern template class distbook<double, distpoint::hist1d>;
  extern template class distbook<std::array<double, 2>, distpoint::hist1d, sample_ratio_traits<std::array<double, 2>>>;
}
    
#endif
