/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/29/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_basic_user_h__
#define __mc_basic_user_h__ 1

//   std includes
#include <iostream>
#include <fstream>
#include <sstream>
#include <functional>
#include <chrono>
#include <iomanip>
#include <tuple>
#include <dlfcn.h>

//   project headers
#include <bits/mc-user.h>
#include <bits/hep-distbook.h>
#include <bits/conf-deductor.h>
#include <cxx/filesystem.h>



namespace duct {
  

  /*
   *   Abstract class for defining basic_user types
   */
  class basic_user_base : public virtual user_base
  {
  public:
    /*   useful type    */
    using duration_type = std::chrono::duration<long double, std::ratio<60,1>>;
    
    /*   destructor   */
    virtual ~basic_user_base() = default;
    
    /*   finish function   */
    void finish();
    
    /*    Returns a function that can read the output of the basic_user<...> class
     *    and sum them up and print the result into file(s) in a nice way.
     */
    virtual std::function<unsigned long long(const fs::path&, const std::list<fs::path>&)>
    collect_and_print_result() const = 0;

    /*   retruns the name of the user   */
    const std::string& name() const { return _M_uname;}
    
    /*   setting the name  of the user   */
    basic_user_base& name(const std::string& uname) {
      _M_uname = uname;
      return *this;
    }
    
    /*   setting the directory of the output file   */
    basic_user_base& dirname(const fs::path& dir) {
      _M_dname = dir;
      return *this;
    }
    
    /*   setting the name of the run   */
    basic_user_base& runname(const char *run) {
      _M_rname = std::string{run};
      return *this;
    }
    
    /*   receives the PID and the name of the deductor process.   */
    void process_info(unsigned int, const char *name) final {
      _M_pname = std::string{name};
    }

    /*   setting some parameters   */
    basic_user_base& param(const duration_type& nsave = duration_type{2}, unsigned long long nevent = 0UL)
    {
      _M_save_period = nsave;
      user_base::max_number_of_events = nevent;
      return *this;
    }
   
    /*   setting the name of the user and the parameters   */
    basic_user_base& param(const char *uname, const duration_type& nsave = duration_type{2}, unsigned long long nevent = 0UL)
    {
      _M_uname = std::string{uname};
      _M_save_period = nsave;
      user_base::max_number_of_events = nevent;
      return *this;
    }
    
    /*   This a simple callback function for do some global initialization of the basic_user classes.   */
    thread_local static std::function<void(basic_user_base *)> initializer;
    
  protected:
    /*    default constructor is protected  */
  basic_user_base() : _M_dname("./"), _M_rname("run"), _M_pname("proc"),
  _M_timer{std::chrono::system_clock::now()}, _M_save_period{duration_type{2}}
  {
    std::ostringstream ss;
    ss<<std::hex<<this;
    _M_uname = ss.str();
      
    /*   global intialization   */
    if(initializer != nullptr) initializer(this);
  }
    
  private:
    /*    operations at the end of the event and saving the results   */
    void operations_at_the_end_of_event() final;

    /*    do any operetaions at the end of the event   */
    virtual void basic_user_end_of_event() = 0;
    
    /*   save the results in txt and binary mode   */
    virtual void write(std::ostream&) = 0;
 
    /*   Returns thee name of the underlying collect_and_print() function in the object code.  */
    virtual const char * c_name_of_collect_and_print() const = 0;

  private:
    /*    data members   */
    fs::path _M_dname;
    std::string _M_uname, _M_rname, _M_pname;

    /*   Assembles the complete output file name   */
    fs::path _M_getfname(bool) const;
  
    /*    time point for statring the calculation   */
    std::chrono::time_point<std::chrono::system_clock> _M_timer;
    
    /*   Time between two saves of the result   */
    duration_type _M_save_period;
  };



  /*   template struct to print latex plot templates    */
  template<class _Distbook>
  class latex_summary
  {
    /*    private types   */
    using _Tp = typename _Distbook::sample_type;
    using _Point = typename _Distbook::point_type;
    using _Traits = typename _Distbook::sample_traits;
    using _PointTraits = typename _Distbook::point_traits;
    using _Iterator = typename _Distbook::const_iterator;
    
  public:
    /*    print the latex template    */
    static void plot(fs::path dir, const std::string& plotdir, const _Distbook& db)
    {
      constexpr char xcol = std::is_same<_Point, distpoint::hist1d>::value ? '1' : '0';
      constexpr int nvar = _Traits::number_of_vars;
      constexpr int c1 = std::is_same<_Point, distpoint::hist1d>::value ? 3 : 1;

      /*   the directory of the plot files   */
      dir /= plotdir;
      
      /*   file of the latex templates   */
      std::fstream ltx((dir/"latex-template.tex").c_str(), std::ios::out);
      
      for(auto&& vi : db)
      {
        /*    writting the latex template file   */
        ltx<<"\n\\subsubsection*{\\it "<<std::get<0>(vi.second)<<"}\n\n"
        <<"\\begin{tikzpicture}\n"
        "\\begin{axis}[title = {"<<std::get<0>(vi.second)<<"},\n"
        "   xlabel={-- x label --}, ylabel={-- y label --},\n"
        "%  xmin=, xmax=,\n"
        "%  ymin=, ymax=\n]\n"
        "\\errorband[red,semithick]{fill=red!30!white, opacity=0.5}{\\mypath/"<<plotdir<<"/table-"<<vi.first<<".dat}{"
        <<xcol<<"}{"<<c1<<"}{"<<c1+1<<"}\n\n";
        
        for(int i = 1; i < nvar; ++i)
          ltx<<"%\\errorband[blue,semithick]{fill=blue!30!white, opacity=0.5}{\\mypath/"
          <<plotdir<<"/table-"<<vi.first<<".dat}{"<<xcol<<"}{"<<c1+2*i<<"}{"<<c1+1+2*i<<"}\n";
        
        ltx<<
        "\\end{axis}\n"
        "\\end{tikzpicture}\n\n";
        
        /*   every table goes into separate file   */
        std::fstream f((dir/(std::string{"table-"} + std::to_string(vi.first)+".dat")).c_str(), std::ios::out);
        f.setf(std::ios::scientific, std::ios::floatfield);
        
        f<<"%  "<<std::get<0>(vi.second)<<"\n";
        ::duct::print(f, std::get<1>(vi.second))<<"\n";
        
        /*   closing the table file   */
        f.flush();
        f.close();
      }
      
      /*     closing the latex template file   */
      ltx.flush();
      ltx.close();
    }
  };


  template<typename _Tp, class _Traits>
  class latex_summary<distbook<_Tp, void, _Traits>>
  {
    /*    private types   */
    using _Distbook = distbook<_Tp, void, _Traits>;
    using _Point = typename _Distbook::point_type;
    using _PointTraits = typename _Distbook::point_traits;
    using _Iterator = typename _Distbook::const_iterator;
    
    static std::string sci(double x, double err)
    {
      std::ostringstream str;
      str.setf(std::ios::scientific, std::ios::floatfield);
      
      str<<"\\pgfmathprintnumber[fixed zerofill,precision=4]{"
      <<x<<"} \\pm \\pgfmathprintnumber[fixed zerofill,precision=4]{"
      <<err<<"}";
      
      return str.str();
    }
    
  public:
    /*     print the latex templates    */
    static void plot(fs::path dir, const std::string& plotdir, const _Distbook& db)
    {
      /*   the directory of the plot files   */
      dir /= plotdir;
      
      /*   file of the latex templates   */
      std::fstream ltx((dir/"latex-template.tex").c_str(), std::ios::out);
      
      for(auto&& vi : db)
      {
        /*    writing the latex template file    */
        ltx<<"\n\\subsubsection*{\\it "<<std::get<0>(vi.second)<<"}\n\n"
        <<"\\begin{equation}\n"
        <<"\\sigma = \\left[\n"
        <<"\\input{\\mypath/"<<plotdir<<"/array-"<<vi.first<<".tex}\n"
        <<"\\right]\n"
        <<"\\end{equation}\n\n";

        /*   every array goes into separate file   */
        std::fstream f((dir/(std::string{"array-"} + std::to_string(vi.first)+".tex")).c_str(), std::ios::out);
        
        /*   writting the array files   */
        auto& acc = std::get<1>(vi.second).get_accumulator();
        std::array<typename _Traits::float_type, _Traits::number_of_vars> mean{acc.mean()}, err{acc.error()};
        
        f<<"\\begin{array}{c}\n";
        for(std::size_t k = 0; k < _Traits::number_of_vars; ++k)
          f<<sci(mean[k], err[k])<<(k+1 < _Traits::number_of_vars ? "\\\\ \n" : "\n");
        f<<"\\end{array}\n";
        
        /*     closing the array file   */
        f.flush();
        f.close();
      }
      
      /*     closing the latex template file   */
      ltx.flush();
      ltx.close();
    }
  };



  /*
   *   Helper class to produce results
   */
  template<class ..._Distbook>
  class __basic_user_make_result
  {
    template<class _Tuple, std::size_t N>
    struct unpacker
    {
      /*   ass_add (+=) operator */
      static void ass_add(_Tuple& r, const _Tuple& x) {
        unpacker<_Tuple, N-1>::ass_add(r, x);
        std::get<N-1>(r) += std::get<N-1>(x);
      }
      
      /*   binary read from a stream   */
      static void read(std::istream& is, _Tuple& x) {
        unpacker<_Tuple, N-1>::read(is, x);
        duct::read(is, std::get<N-1>(x));
      }
      
      /*   print the reults into files   */
      static void print(unsigned long long nsaved, fs::path fn, const _Tuple& x)
      {
        unpacker<_Tuple, N-1>::print(nsaved, fn, x);
        
        /*   printing every distbook into to separate files   */
        fn /= std::string{"result-"}+std::to_string(N-1);
        
        /*  printing   */
        std::fstream f(fn.c_str(), std::ios::out);
        std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
        
        f<<"#  This result was calculated by Deductor version "<<conf::version<<".\n"
        << "#  The number of collected MC events is "<<nsaved<<".\n"
        << "#  It was created on ";
        char mbstr[100];
       
        if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n\n", std::localtime(&now)))
          f<<mbstr;
      
        
        f.setf(std::ios::scientific, std::ios::floatfield);
        ::duct::print(f, std::get<N-1>(x));
        f.close();
      }
      
      /*   plot the result into directories  */
      static void plot(unsigned long long nsaved, fs::path fn, const _Tuple& x)
      {
        unpacker<_Tuple, N-1>::plot(nsaved, fn, x);
       
        /*   every distbook goes to separate directory   */
        std::string plotdir{"plot-" + std::to_string(N-1)};
        
        fs::create_directories(fn/plotdir);
        latex_summary<typename std::tuple_element<N-1, _Tuple>::type>::plot(fn, plotdir, std::get<N-1>(x));
      }
    };
    
  public:
    /*   result type is a tuple   */
    typedef std::tuple<_Distbook...> result_type;
    
    /*   reading from stream   */
    static unsigned long long read(std::istream& is, result_type& x)
    {
      unsigned long long nsaved;
      std::string dummy;
      
      /*   reading the file   */
      duct::read(is, dummy, dummy, nsaved);
      unpacker<result_type, sizeof...(_Distbook)>::read(is, x);
      
      return nsaved;
    }
 
    /*   collect the files and print the result   */
    static unsigned long long collect_and_print(const fs::path&, const std::list<fs::path>&);
  };
  
  
  template<class ..._Distbook> template<class _Tuple>
  struct __basic_user_make_result<_Distbook...>::unpacker<_Tuple, 1>
  {
    /*   ass_add (+=) operator */
    static void ass_add(_Tuple& r, const _Tuple& x) {
      std::get<0>(r) += std::get<0>(x);
    }
    
    /*   binary read from a stream   */
    static void read(std::istream& is, _Tuple& x) {
      ::duct::read(is, std::get<0>(x));
    }
    
    /*   print the reults into files   */
    static void print(unsigned long long nsaved, fs::path fn, const _Tuple& x)
    {
      /*   printing every distbook into to separate files   */
      fn /= std::string{"result-0"};
      
      /*  printing   */
      std::fstream f(fn.c_str(), std::ios::out);
      std::time_t now = std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());
      char mbstr[100];
     
      f<<"#  This result was calculated by Deductor version "<<conf::version<<".\n"
      << "#  The number of collected MC events is "<<nsaved<<".\n"
      << "#  It was created on ";

      if(std::strftime(mbstr, 100, "%A, %Y %B %e at %T %Z.\n\n", std::localtime(&now)))
        f<<mbstr;
      
      f.setf(std::ios::scientific, std::ios::floatfield);
      ::duct::print(f, std::get<0>(x));
      f.close();
    }
    
    /*   plot the result into directories  */
    static void plot(unsigned long long nsaved, fs::path fn, const _Tuple& x)
    {
      /*   every distbook goes to separate directory   */
      std::string plotdir{"plot-0"};
      
      fs::create_directories(fn/plotdir);
      latex_summary<typename std::tuple_element<0, _Tuple>::type>::plot(fn, plotdir, std::get<0>(x));
    }
  };

  
  template<class ..._Distbook> unsigned long long int
  __basic_user_make_result<_Distbook...>::collect_and_print(const fs::path& dir, const std::list<fs::path>& files)
  {
    std::fstream file;
    unsigned long int nsaved = 0UL;
    result_type tmp, *res = nullptr;
    
    for(auto& fi : files) {
      //----- open the file ----
      file.open(fi.c_str(), std::ios::in);
      
      //----- read the file if it is exist -----
      if(file.is_open()) {
        nsaved += read(file, tmp);
        file.flush();
        
        if(!res) res = new result_type(tmp);
        else unpacker<result_type, sizeof...(_Distbook)>::ass_add(*res, tmp);
      }
      
      //----- close the file ----
      file.close();
    }
    
    if(!res) return 0;
    
    /*    create the result directory   */
    fs::create_directories(dir);
    
    //----- save the results into the directory dir  -----
    unpacker<result_type, sizeof...(_Distbook)>::print(nsaved, dir, *res);
    unpacker<result_type, sizeof...(_Distbook)>::plot(nsaved, dir, *res);
    
    //---- cleaning up -----
    if(res) delete res;
    
    /*   analyser info   */
    std::fstream ltx((dir/"analyser-info.tex").c_str(), std::ios::out);
    
    ltx<<"\\def\\mcevent{"<<nsaved<<"}\n";
    
    ltx.flush();
    ltx.close();
    
    return nsaved;
  } 
  

  /*   This is a wrapper class around distbook. It provides the functinality
   *   that is needed in the basic_user classes in an analyser.
   */
  template<class _Distbook>
  class __distbook_wrapper
  {
    /*    private types   */
    using _Sample = typename _Distbook::sample_type;
    using _Point = typename _Distbook::point_type;
 
  protected:
    /*    create an entry in the book   */
    void phys(int key, const char *title, unsigned n, const _Point *base, bool norm = false) {
      _M_book.create(key, title, n, base, norm);
    }
    
    template<class _InputIterator>
    void phys(int key, const char *title, _InputIterator first, _InputIterator last, bool norm = false) {
      _M_book.create(key, title, first, last, norm);
    }
    
    void phys(int key, const char *title, std::initializer_list<_Point> base, bool norm = false) {
      _M_book.create(key, title, base, norm);
    }
    
    template<class _Container>
    void phys(int key, const char *title, const _Container& base, bool norm = false) {
      _M_book.create(key, title, base, norm);
    }
    
    /*   collecting weight   */
    template<class _Func>
    void physfill(int key, _Func&& func) {
      _M_book.collect(key, std::forward<_Func>(func));
    }
    
    template<class _Func, typename ..._Arg>
    void physfill(int key, const _Sample& w, _Func&& func, _Arg&&... x) {
      _M_book.collect(key, w, std::forward<_Func>(func), std::forward<_Arg>(x)...);
    }
    
    /*   collects weights for the normalization   */
    void normfill(int key, const _Sample& w) {
      _M_book.collect_norm(key, w);
    }
    
  private:
    //    data member
    _Distbook _M_book;

    /*   friend declaration   */
    template<class...> friend class __distbook_pack;
  };
  
  
  template<typename _Sample, class _Traits>
  class __distbook_wrapper<distbook<_Sample, void, _Traits, distpoint::traits<void>>>
  {
  protected:
    /*    create an entry in the book   */
    void phys(int key, const char *title) { _M_book.create(key, title);}
    
    /*   collecting weight   */
    void physfill(int key, const _Sample& w) { _M_book.collect(key, w);}
    
  private:
    /*    data members  */
    distbook<_Sample, void, _Traits, distpoint::traits<void>> _M_book;
    
    /*   friend declaration   */
    template<class...> friend class __distbook_pack;
  };


  
  template<class ..._DistBook>
  struct __distbook_pack : public __distbook_wrapper<_DistBook>...
  {
    /*   helper class to unpack the template variable   */
    template<class _Arg, class ..._Args>
    struct unpacker
    {
      static void end_of_event(__distbook_pack *bs) {
        bs->__distbook_wrapper<_Arg>::_M_book.accumulate();
        unpacker<_Args...>::end_of_event(bs);
      }
      
      static void write(__distbook_pack *bs, std::ostream& os) {
        ::duct::write(os, bs->__distbook_wrapper<_Arg>::_M_book);
        unpacker<_Args...>::write(bs, os);
      }
    };
  };
  
  
  template<class ..._DistBook> template<class _Arg>
  struct __distbook_pack<_DistBook...>::unpacker<_Arg>
  {
    static void end_of_event(__distbook_pack<_DistBook...> *bs) {
      bs->__distbook_wrapper<_Arg>::_M_book.accumulate();
    }
    
    static void write(__distbook_pack<_DistBook...> *bs, std::ostream& os) {
      ::duct::write(os, bs->__distbook_wrapper<_Arg>::_M_book);
    }
  };

  
  
  template<class _Tag, class ..._DistBook>
  class basic_user : public basic_user_base, public user<_Tag>, protected __distbook_pack<_DistBook...>
  {
  public:
    /*   result generator   */
    std::function<unsigned long long(const fs::path&, const std::list<fs::path>&)>
    collect_and_print_result() const final {
      return __basic_user_make_result<_DistBook...>::collect_and_print;
    }
    
    /*   static version of result generator   */
    static unsigned long long collect_and_print(const fs::path& dir, const std::list<fs::path>& files) {
      return __basic_user_make_result<_DistBook...>::collect_and_print(dir, files);
    }
    
    /*   Aliases to the distbook types in the template arguments.   */
    template<std::size_t _Nth>
    using distbook_t = typename std::tuple_element<_Nth, std::tuple<_DistBook...>>::type;

    template<class T>
    using Phys = __distbook_wrapper<T>;

  private:
    /*   overriding the virtual function   */
    void basic_user_end_of_event()   final { __distbook_pack<_DistBook...>::template unpacker<_DistBook...>::end_of_event(this);}
    void write(std::ostream& os) final { __distbook_pack<_DistBook...>::template unpacker<_DistBook...>::write(this, os);}
    
    const char * c_name_of_collect_and_print() const final {
      ::Dl_info info;
      if(::dladdr(reinterpret_cast<void *>(&__basic_user_make_result<_DistBook...>::collect_and_print), &info))
        return info.dli_sname;
      return nullptr;
    }
  };
  
}  //   namespace duct

#endif
