/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 8/19/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_deductor_h__
#define __mc_deductor_h__ 1

/*   std includes    */
#include <cstdlib>
#include <list>
#include <vector>
#include <queue>
#include <string>
#include <thread>
#include <mutex>
#include <chrono>
#include <atomic>
#include <type_traits>
#include <condition_variable>

/*   project headers    */
#include <bits/duct-tags.h>
#include <bits/mc-user.h>
#include <bits/mc-engine.h>
#include <bits/mc-hard-state.h>
#include <bits/mc-ordering.h>
#include <bits/mc-splitting-prob.h>



namespace duct {
  namespace __details {

    /*    This class is to store the buffer (a set of shower_history objects) and 
     *    provides functions to fill with real event and analyse them.
     */
    template<typename _Tag>
    class deductor_buffer
    {
    public:
      /*   Create a buffer with n elements   */
      explicit deductor_buffer(std::size_t n = 32u)
      : _M_pool(n), _M_kill{false}, _M_newusr{nullptr}
      { for(std::size_t k = 0; k < n; k++) _M_consumed.push(k);}
      
      /*   Copy and move are deleted because the mutex and conditional_variable
       *   cannot be copied or moved.
       */
      deductor_buffer(const deductor_buffer&) = delete;
      deductor_buffer& operator=(const deductor_buffer&) = delete;
      
      deductor_buffer(deductor_buffer&&) = delete;
      deductor_buffer& operator=(deductor_buffer&&) = delete;
      
      ~deductor_buffer() = default;
      
     
      /*  This is the shower generator. It write new events to the buffer.
       *  It stops when it receives kill signal form the analyzer routine.
       */
      void generator(hard_state<_Tag>&);
      void generator(hard_state<_Tag>&, shower_engine<_Tag>&);
      
      /*  This is the analyser routine. It reads elements from a buffer and call the user
       *  routines. It stops and send the kill signal when all the analysises are done.
       */
      void analyser(std::list<std::unique_ptr<user<_Tag>>> *);
      
      /*   Checked if the calculation is finished   */
      bool is_killed() const { return _M_kill;}
      
      /*   Send the kill signal too all of the threads   */
      void kill() { _M_kill = true;}
      
      /*   Attach a new user routine  (the user has to be initialized) */
      void attach_usr(user<_Tag> *usr) {
        _M_newusr = usr;
        _M_user_is_not_empty.notify_one();
      }
      
    private:
      /*   This is the pool where we store the elements of the buffer   */
      std::vector<shower_history<_Tag>> _M_pool;
      
      /*    This is the queue of the produced objects.  */
      std::queue<std::size_t> _M_produced;
      
      /*   This is the queue of the consumed objects. They are ready for recycling.  */
      std::queue<std::size_t> _M_consumed;
      
      /*   Signal to kill the generators   */
      std::atomic<bool> _M_kill;
      std::atomic<user<_Tag> *> _M_newusr;
      
      /*    Mutexes for protecting the queues.   */
      std::mutex _M_pmtx, _M_cmtx, _M_umtx;
      
      /*   Conditional variables to syncronize threads    */
      std::condition_variable _M_prod_is_not_empty;
      std::condition_variable _M_cons_is_not_empty;
      std::condition_variable _M_user_is_not_empty;
    };
    
    /*   Function to generate the event. We can have several generator task. They can run 
     *   simultaneously and deductor_buffer class takes care about the syncronisation.
     */
    template<class _Tag>
    void deductor_buffer<_Tag>::generator(hard_state<_Tag>& hard, shower_engine<_Tag>& engine)
    {
      /*     initialize the random number generator   */
      std::size_t seed = std::hash<std::thread::id>()(std::this_thread::get_id());
      seed ^= std::hash<decltype(this)>()(this) << 1;
      seed ^= std::hash<unsigned long>()(std::chrono::system_clock::to_time_t(std::chrono::system_clock::now())) << 1;
      
      std::size_t seedH = seed ^ (std::hash<hard_state<_Tag>*>()(&hard) << 1);
      std::size_t seedE = seed ^ (std::hash<shower_engine<_Tag>*>()(&engine) << 1);
      
      hard.seed(seedH);
      engine.seed(seedE);
         

      /*   The generator runs until it gets the kill signal.   */
      while(_M_kill == false) {
        /*   pop from the consumed queue   */
        unsigned int current;
        //------
        {
          std::unique_lock<std::mutex> lock(_M_cmtx);
          while(_M_consumed.empty()) {
            if(_M_kill) return;
            _M_cons_is_not_empty.wait(lock);
          }
          
          current = _M_consumed.front(); _M_consumed.pop();
        }
        
        /*   Generate the hard event and do the shower evolution.   */
        shower_history<_Tag> &hist = _M_pool[current];
        
        /*   generate the hard event(s)   */
        hard(hist);
        
        /*   do the shower evolution   */
        engine(1e+120, hist);
        
        /*   puts the product into the produced queue   */
        _M_pmtx.lock(); _M_produced.push(current); _M_pmtx.unlock();
        _M_prod_is_not_empty.notify_one();
      }
    }
    
    
    template<class _Tag>
    void deductor_buffer<_Tag>::generator(hard_state<_Tag>& hard)
    {
      /*     initialize the random number generator   */
      std::size_t seed = std::hash<std::thread::id>()(std::this_thread::get_id());
      seed ^= std::hash<decltype(this)>()(this) << 1;
      seed ^= std::hash<unsigned long>()(std::chrono::system_clock::to_time_t(std::chrono::system_clock::now())) << 1;
      
      std::size_t seedH = seed ^ (std::hash<hard_state<_Tag>*>()(&hard) << 1);
      hard.seed(seedH);
      
      
      /*   The generator runs until it gets the kill signal.   */
      while(_M_kill == false) {
        /*   pop from the consumed queue   */
        unsigned int current;
        //------
        {
          std::unique_lock<std::mutex> lock(_M_cmtx);
          while(_M_consumed.empty()) {
            if(_M_kill) return;
            _M_cons_is_not_empty.wait(lock);
          }
          
          current = _M_consumed.front(); _M_consumed.pop();
        }
        
        /*   Generate the hard event and do the shower evolution.   */
        shower_history<_Tag> &hist = _M_pool[current];
        
        /*   generate the hard event(s)   */
        hard(hist);
        
        /*   puts the product into the produced queue   */
        _M_pmtx.lock(); _M_produced.push(current); _M_pmtx.unlock();
        _M_prod_is_not_empty.notify_one();
      }
    }
    
    
    /*   This function runs the analyser routines. It takes user functions and run them
     *   sequentially on every shower histories.
     */
    template<class _Tag>
    void deductor_buffer<_Tag>::analyser(std::list<std::unique_ptr<user<_Tag>>> *users)
    {
      /*   This is a fresh calculation, let the generators run...  */
      _M_kill = false;
      
      /*   The analyzer runs until all the user routines are done.   */
      while(_M_kill == false) {
        /*   Check if there is a new user to attachment   */
        user<_Tag> *newusr = _M_newusr.exchange(nullptr);
        if(newusr) users->push_back(std::unique_ptr<user<_Tag>>(newusr));
        
        /*   If all the analyseses are done, wait 20 minutes for new user.
         *   If there is no new attachment send the kill signal.
         */
        if(users->empty())
        {
          std::unique_lock<std::mutex> lock(_M_umtx);
          user<_Tag> *unew = _M_newusr.exchange(nullptr);
          
          if(unew) users->push_back(std::unique_ptr<user<_Tag>>(unew));
          else {
            if(_M_user_is_not_empty.wait_for(lock, std::chrono::minutes(10)) == std::cv_status::no_timeout)
              users->push_back(std::unique_ptr<user<_Tag>>(_M_newusr.exchange(nullptr)));
            else {
              _M_kill = true;
              break;
            }
          }
        }
        
        /*   pop from the produced queue   */
        unsigned int current;
        //------
        {
          std::unique_lock<std::mutex> lock(_M_pmtx);
          while(_M_produced.empty()) _M_prod_is_not_empty.wait(lock);
          current = _M_produced.front(); _M_produced.pop();
        }
        
        /*   Analysing the event  */
        const shower_history<_Tag>& history = _M_pool[current];
        
        for(auto iu = users->begin(); iu != users->end();) {
          (*iu)->userfunc(history);
          (*iu)->end_of_event();
          
          if((*iu)->is_done()) {
            (*iu)->finish();
            iu = users->erase(iu);
          } else ++iu;
        }
        
        /*   puts the consumed item back to the consumed queue   */
        _M_cmtx.lock(); _M_consumed.push(current); _M_cmtx.unlock();
        _M_cons_is_not_empty.notify_one();
      }
      
      /*   There was a kill signal, notify all blocked generators to quit.   */
      _M_cons_is_not_empty.notify_all();
    }
    
    
    
    /*    This is a class for representing a deductor process.
     *    It starts many generator and one analyser thread and lets them to run.
     *    The instances of this class own the threads and they join in the
     *    destructor of this class.
     */
    template<class _Tag>
    class deductor_task
    {
    public:
      /*   default constructor  */
      deductor_task() : _M_usr(nullptr), _M_buffer{nullptr} {}
      
      /*   copy is deleted   */
      deductor_task(const deductor_task&) = delete;
      deductor_task& operator=(const deductor_task&) = delete;
      
      /*   move is allowed    */
      deductor_task(deductor_task&& x)
      : _M_usr(nullptr), _M_buffer(nullptr), _M_threads(std::move(x._M_threads))
      { std::swap(_M_usr, x._M_usr); std::swap(_M_buffer, x._M_buffer);}
      
      deductor_task& operator=(deductor_task&& x) {
        std::swap(_M_usr, x._M_usr);
        std::swap(_M_buffer, x._M_buffer);
        _M_threads = std::move(x._M_threads);
        return *this;
      }
      
      /*   destructor  */
      ~deductor_task() {
        /*   Every thread has to be joined to the main thread before
         *   the data members get destroyed.
         */
        for(auto& gi : _M_threads) gi.join();
        
        /*   deleting the buffer   */
        if(_M_buffer) delete _M_buffer;
        
        /*  deleting the user list  */
        if(_M_usr) delete _M_usr;
      }
      
      /*   Creates and starts the calculation for a given hard process and with the user functions.   */
      deductor_task(const hard_state<_Tag>&, std::list<std::unique_ptr<user<_Tag>>>&&, int);
      deductor_task(const hard_state<_Tag>&, const shower_engine<_Tag>&, std::list<std::unique_ptr<user<_Tag>>>&&, int);
      
      /*   Checked if the calculation is finished   */
      bool is_killed() const { return _M_buffer ? _M_buffer->is_killed() : true;}
      
      /*   Send the kill signal too all of the threads   */
      void kill() { if(_M_buffer) _M_buffer->kill();}
      
      /*   Attach a new user routine  */
      void attach_usr(user<_Tag> *usr) {
        usr->initialize();
        _M_buffer->attach_usr(usr);
      }
      
    private:
      /*    List of user functions.  */
      std::list<std::unique_ptr<user<_Tag>>> *_M_usr;
      
      /*   Buffer to queue the generated events.   */
      deductor_buffer<_Tag> *_M_buffer;
      
      /*   Vector of threads   */
      std::vector<std::thread> _M_threads;
    };
    
    
    template<class _Tag> deductor_task<_Tag>::
    deductor_task(const hard_state<_Tag>& hard, const shower_engine<_Tag>& engine, std::list<std::unique_ptr<user<_Tag>>>&& usr, int load)
    {
      /*   move the user list  */
      _M_usr = new std::list<std::unique_ptr<user<_Tag>>>{std::move(usr)};
      
      /*   Initialize the user functions   */
      for(const auto& p : *_M_usr) p->initialize();
      
      /*   if it not specified then the number of thread is the number of available CPUs   */
      unsigned nthr = std::thread::hardware_concurrency();
      unsigned ngen = nthr;
      
      if(load == 0u) ngen = nthr;
      else if(load < 0) ngen = 1 + nthr/std::abs(load);
      else ngen = load;
      
      if(ngen > nthr) ngen = nthr;
      if(ngen == 0u) ngen = 1u;
      
      /*   Buffer to queue the generated events.   */
      _M_buffer = new deductor_buffer<_Tag>{2*ngen+20u};
      
      /*   Vector of threads   */
      _M_threads.resize(ngen+1);
      
      /*   Starting the analyser process   */
      _M_threads[0] = std::thread(&deductor_buffer<_Tag>::analyser, _M_buffer, _M_usr);
      
      /*   Starting the generator process(es)   */
      for(unsigned i = 1; i <= ngen; ++i)
        _M_threads[i] = std::thread([h = hard, e = engine](deductor_buffer<_Tag> *buffer) mutable { buffer -> generator(h, e);}, _M_buffer);
    }
 
    
    template<class _Tag> deductor_task<_Tag>::
    deductor_task(const hard_state<_Tag>& hard, std::list<std::unique_ptr<user<_Tag>>>&& usr, int load)
    {
      /*   move the user list  */
      _M_usr = new std::list<std::unique_ptr<user<_Tag>>>{std::move(usr)};
      
      /*   Initialize the user functions   */
      for(const auto& p : *_M_usr) p->initialize();
      
      /*   if it not specified then the number of thread is the number of available CPUs   */
      unsigned nthr = std::thread::hardware_concurrency();
      unsigned ngen = nthr;
      
      if(load == 0u) ngen = nthr;
      else if(load < 0) ngen = 1 + nthr/std::abs(load);
      else ngen = load;
      
      if(ngen > nthr) ngen = nthr;
      if(ngen == 0u) ngen = 1u;
      
      /*   Buffer to queue the generated events.   */
      _M_buffer = new deductor_buffer<_Tag>{2*ngen+20u};
      
      /*   Vector of threads   */
      _M_threads.resize(ngen+1);
      
      /*   Starting the analyser process   */
      _M_threads[0] = std::thread(&deductor_buffer<_Tag>::analyser, _M_buffer, _M_usr);
      
      /*   Starting the generator process(es)   */
      for(unsigned i = 1; i <= ngen; ++i)
        _M_threads[i] = std::thread([h = hard](deductor_buffer<_Tag> *buffer) mutable { buffer -> generator(h);}, _M_buffer);
    }
    
    
    
    /*    We nneed some runtime data for the running thread.
     *    This has to be alive as long as there is any running thread.
     */
    struct deductor_kernel_rtd
    {
      /*   constructors   */
      deductor_kernel_rtd() : pid{1u}, mdl{} {}
      deductor_kernel_rtd(model&& m) : pid{1u}, mdl{std::move(m)} {}
      deductor_kernel_rtd(const model& m) : pid{1u}, mdl{m} {}
      
      /*   Keeps tracking the process ID   */
      unsigned int pid;
      
      /*   model   */
      model mdl;
    };
    
  }   //   namespace details
  
  
  

  /*   This is the abstract deductor base class. It can handle the processes (start/kill)
   *   but doesn't know anything about the shower engine. 
   */
  template<class _Tag>
  class deductor_kernel : private __details::deductor_kernel_rtd
  {
    /*    some type aliases   */
    using _Rtd = __details::deductor_kernel_rtd;
    using _Task = __details::deductor_task<_Tag>;
    
  public:
    /*   constructors   */
    deductor_kernel() = default;
    explicit deductor_kernel(model&& m) :_Rtd{std::move(m)} {}
    explicit deductor_kernel(const model& m) :_Rtd{m} {}

    /*    destructor   */
    ~deductor_kernel() = default;
    
    
    /*   copy is deleted because the process cannot be copied   */
    deductor_kernel(const deductor_kernel&) = delete;
    deductor_kernel& operator=(const deductor_kernel&) = delete;

    /*   move   */
    deductor_kernel(deductor_kernel&&) = default;
    deductor_kernel& operator=(deductor_kernel&&) = default;

    
    /*   Checked if the process is killed   */
    bool is_killed(unsigned int __pid) const {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end())
        return p->second.second.is_killed();
      return true;
    }
    
    /*    Send a kill signal to a process    */
    void kill(unsigned int __pid) {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end()) {
        p->second.second.kill();
        _M_proc.erase(p);
      }
    }
    
    /*   Attach a new user routine   */
    void attach_usr(unsigned int __pid, user<_Tag> *usr) {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end())
        p->second.second.attach_usr(usr);
    }
    
    /*   returns the name of the process with the given ID    */
    std::string name(unsigned __pid) const {
      auto p = _M_proc.find(__pid);
      if(p != _M_proc.end()) return p->second.first;
      
      return std::string{"PID "} + std::to_string(__pid) + std::string{" not found!"};
    }

    
    /*   Start a shower calculation with a given hard process and a user function(s).   */
    unsigned operator()(const char *, const hard_state<_Tag>&, const shower_engine<_Tag>&, std::list<std::unique_ptr<user<_Tag>>>&&, int);
    
    /*   Start a fixed ordered calculation with a given hard process and a user function(s).   */
    unsigned operator()(const char *, const hard_state<_Tag>&, std::list<std::unique_ptr<user<_Tag>>>&&, int);
    
    /*   get the model   */
    const model * get_model() const { return std::addressof(_Rtd::mdl);}
    
  private:
    /*    Process table   */
    std::map<unsigned, std::pair<std::string, _Task>> _M_proc;
  };
  
  
  template<class _Tag> unsigned int deductor_kernel<_Tag>::
  operator()(const char *name, const hard_state<_Tag>& hard, const shower_engine<_Tag>& engine, std::list<std::unique_ptr<user<_Tag>>>&& usr, int ngen)
  {
    /*   The PID of this process   */
    unsigned int __pid = _Rtd::pid++;
    
    for(const auto& u : usr)
      u->process_info(__pid, name);
 
     /*   Inserting the new process    */
    _M_proc[__pid] = {name, _Task{hard, engine, std::move(usr), ngen}};
    
    return __pid;
  }
  
  template<class _Tag> unsigned int deductor_kernel<_Tag>::
  operator()(const char *name, const hard_state<_Tag>& hard, std::list<std::unique_ptr<user<_Tag>>>&& usr, int ngen)
  {
    /*   The PID of this process   */
    unsigned int __pid = _Rtd::pid++;
    
    for(const auto& u : usr)
      u->process_info(__pid, name);
    
    /*   Inserting the new process    */
    _M_proc[__pid] = {name, _Task{hard, std::move(usr), ngen}};
    
    return __pid;
  }
  

  

  namespace __details {
    
    /*   Class deductor to generate and analyse events.
     *   This is the high level interface what most user really want.
     */
    template<class _Prob>
    class deductor_impl : public deductor_kernel<typename _Prob::process_type>
    {
      using _Kernel = deductor_kernel<typename _Prob::process_type>;
      using _Engine = shower_engine<typename _Prob::process_type>;
      using _HardState = hard_state<typename _Prob::process_type>;
      
      using _UsrPtr = std::unique_ptr<user<typename _Prob::process_type>>;
      using _UsrList = std::list<_UsrPtr>;
      
    public:
      /*   public types  */
      using process_type = typename _Prob::process_type;
      using ordering_type = typename _Prob::ordering_type;
      
      
      /*   default constructor   */
      deductor_impl() = default;
      
      /*   create from model   */
      explicit deductor_impl(model&& mdl) : _Kernel{std::forward<model>(mdl)}, _M_engine{_Prob{}} {};
      explicit deductor_impl(const model& mdl) : _Kernel{mdl}, _M_engine{_Prob{}} {};
      explicit deductor_impl(const qcd::input& qin) : deductor_impl{model{ordering_type{}, qin}} {}
      explicit deductor_impl(const qcd::predefined& qin) : deductor_impl{model{ordering_type{}, qin}} {}

      /*   the first argument is the model dir    */
      deductor_impl(const fs::path& dir, const qcd::input& qin)
      : deductor_impl{model{dir, ordering_type{}, qin}} {}
      
      
      /*   Constructor with the argument of the underlying shower probability   */
      template<typename ..._Args>
      explicit deductor_impl(const model& mdl, _Args&& ...args)
      : _Kernel{mdl}, _M_engine{_Prob{std::forward<_Args>(args)...}} {}
      
      template<typename ..._Args>
      explicit deductor_impl(model&& mdl, _Args&& ...args)
      : _Kernel{std::forward<model>(mdl)}, _M_engine{_Prob{std::forward<_Args>(args)...}} {}
      
      template<typename ..._Args>
      explicit deductor_impl(const qcd::input& qin, _Args&& ...args)
      : _Kernel{model{ordering_type{}, qin}}, _M_engine{_Prob{std::forward<_Args>(args)...}} {}
      
      /*   the first argument is the model dir    */
      template<typename ..._Args>
      deductor_impl(const fs::path& dir, const qcd::input& qin, _Args&& ...args)
      : _Kernel{model{dir, ordering_type{}, qin}}, _M_engine{_Prob{std::forward<_Args>(args)...}} {}
      
      
      /*   copy is deleted because the process cannot be copied   */
      deductor_impl(const deductor_impl&) = delete;
      deductor_impl& operator=(const deductor_impl&) = delete;
      
      /*   move   */
      deductor_impl(deductor_impl&&) = default;
      deductor_impl& operator=(deductor_impl&&) = default;
      
      /*   destructor   */
      ~deductor_impl() = default;
      
      
      /*   Start a calculation: We have four different ways to do this.      */
      template<class _Hard>
      unsigned operator()(const char *name, _Hard&& mkh, user<process_type> *usr, int ngen = 0u)
      {
        _UsrList users;
        users.push_back(_UsrPtr{usr});
        
        return _Kernel::operator()(name,
                                   _HardState{mkh(model_ref{ordering_type{}, _Kernel::get_model()})},
                                   _M_engine,
                                   std::move(users),
                                   ngen);
      }
      
      
      template<class _Hard>
      unsigned operator()(const char *name, const shower_engine_opts& opts, _Hard&& mkh, user<process_type> *usr, int ngen = 0u)
      {
        /*   preparing engine   */
        shower_engine_opts old_opts = _M_engine.options();
        _M_engine.options(opts);
        
        /*   preparing user list   */
        _UsrList users;
        users.push_back(_UsrPtr{usr});
        
        /*    start process   */
        unsigned pid = _Kernel::operator()(name,
                                           _HardState{mkh(model_ref{ordering_type{}, _Kernel::get_model()})},
                                           _M_engine,
                                           std::move(users),
                                           ngen);
        
        /*   reset engine   */
        _M_engine.options(old_opts);
        
        return pid;
      }
      
      
      template<class _Hard>
      unsigned operator()(const char *name, _Hard&& mkh, std::initializer_list<user<process_type>*>  usr, int ngen = 0u)
      {
        /*   preparing user list   */
        _UsrList users;
        
        for(user<process_type> *u : usr)
          users.push_back(_UsrPtr{u});
        
        /*    start process   */
        return _Kernel::operator()(name,
                                   _HardState{mkh(model_ref{ordering_type{}, _Kernel::get_model()})},
                                   _M_engine,
                                   std::move(users),
                                   ngen);
      }
 

      template<class _Hard>
      unsigned operator()(const char *name, const shower_engine_opts& opts, _Hard&& mkh, std::initializer_list<user<process_type>*>  usr, int ngen = 0u)
      {
        /*   preparing engine   */
        shower_engine_opts old_opts = _M_engine.options();
        _M_engine.options(opts);
        
        /*   preparing user list   */
        _UsrList users;
        
        for(user<process_type> *u : usr)
          users.push_back(_UsrPtr{u});
        
        /*    start process   */
        unsigned pid = _Kernel::operator()(name,
                                           _HardState{mkh(model_ref{ordering_type{}, _Kernel::get_model()})},
                                           _M_engine,
                                           std::move(users),
                                           ngen);
        
        /*   reset engine   */
        _M_engine.options(old_opts);
        
        return pid;
      }
      
      
      /*   Start a fixed order calculation: We have two different ways to do this.      */
      template<class _Hard>
      unsigned fixed_order(const char *name, _Hard&& mkh, user<process_type> *usr, int ngen = 0u)
      {
        _UsrList users;
        users.push_back(_UsrPtr{usr});
        
        return _Kernel::operator()(name,
                                   _HardState{mkh(model_ref{ordering_type{}, _Kernel::get_model()})},
                                   std::move(users),
                                   ngen);
      }
      
      template<class _Hard>
      unsigned fixed_order(const char *name, _Hard&& mkh, std::initializer_list<user<process_type>*>  usr, int ngen = 0u)
      {
        /*   preparing user list   */
        _UsrList users;
        
        for(user<process_type> *u : usr)
          users.push_back(_UsrPtr{u});
        
        /*    start process   */
        return _Kernel::operator()(name,
                                   _HardState{mkh(model_ref{ordering_type{}, _Kernel::get_model()})},
                                   std::move(users),
                                   ngen);
      }
      
   
      /*   Setting/getting the engine options    */
      void options(const shower_engine_opts& opts) { _M_engine.options(opts);}
      const shower_engine_opts& options() const { return _M_engine.options();}
      
    private:
      /*   The prototype of the shower engine. This is copied in the multi-threaded version.
       *   In general the shower engine can have large runtime data. This is the reason why
       *   we keep a prototype.
       */
      _Engine _M_engine;
    };
    
  }   //   namespace __details
  
  
  
  
  /*   The high level user interface of the shower generator   */
  template<class _Proc, class _Ordering = ordering::lambda>
  using deductor = __details::deductor_impl<splitting_prob<_Proc, _Ordering>>;
  
  
  
  /*   Explicit instatiation for reducing compilation time   */
  extern template class  __details::deductor_impl<splitting_prob<epa, ordering::kT>>;
  extern template class  __details::deductor_impl<splitting_prob<dis, ordering::kT>>;
  extern template class  __details::deductor_impl<splitting_prob<hhc, ordering::kT>>;

  extern template class  __details::deductor_impl<splitting_prob<epa, ordering::lambda>>;
  extern template class  __details::deductor_impl<splitting_prob<dis, ordering::lambda>>;
  extern template class  __details::deductor_impl<splitting_prob<hhc, ordering::lambda>>;
}

#endif
