/*  Copyright (C) 2013 Zoltán Nagy
 *  Created by Zoltán Nagy on 4/24/13.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_engine_h__
#define __mc_engine_h__ 1



/*   std includes   */
#include <cmath>
#include <utility>
#include <fstream>
#include <string>
#include <vector>


/*   project headers   */
#include <bits/mc-shower-history.h>
#include <bits/mc-hard-params.h>
#include <bits/mc-splitting-prob.h>
#include <bits/qcd-color-overlap.h>
#include <cxx/cloned-ptr.h>



namespace duct {
  
  namespace __details {
    
    /*  This function selects the color operator. This can be different in different
     *  shower implementations. This function select the color operator in dipole shower.
     */
    double selecting_color_operator_su3(random_engine_type&, splitting_variables<shower<lo>> *, const particle_state *, const model_ref&);
    double selecting_color_operator_u3(random_engine_type&, splitting_variables<shower<lo>> *, const particle_state *, const model_ref&);
    
    
    template<class _Proc> inline
    double selecting_color_operator_su3(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_su3(rne, &(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
    
    template<class _Proc> inline
    double selecting_color_operator_u3(random_engine_type& rne, shower_stage<_Proc>& stage) {
      auto qcd = stage.state.template data<qcd_out>();
      return selecting_color_operator_u3(rne, &(stage.vars), qcd.begin()-1, stage.state.get_model());
    }
    
  }    //   namespace __details

  
  
  /*   Simple structure to provide input parameters for the shower engine.   */
  struct shower_engine_opts
  {
    /*   cutoff kT scale   */
    double kT_cutoff = 1.0_GeV;
    
    /*   max allowed color suppression   */
    unsigned max_color_suppression = 0u;
    
    /*   calculate threshold weight   */
    struct {
      /*   enable or disable threshold weights   */
      bool is_enabled = true;
    } threshold;
  };
  
  
  /*  This is the shower evolution code. It uses a given shower probability class to make 
   *  trial emission. An instance of the shower probability class is stored in the this
   *  object. The type of the probability class is erased.
   */
  template<class _Proc>
  class shower_engine
  {
    template<typename T> static constexpr bool
    __non_self = (!std::is_same<shower_engine, typename std::decay<T>::type>::value
                  && !std::is_base_of<shower_engine, typename std::decay<T>::type>::value);
   
  public:
    /*    default constructor    */
    shower_engine() { _M_not_showered.reserve(64);}
    
    /*    create the shower engine from a given shower probability   */
    template<class _ShowerProb, typename _X = typename std::enable_if_t<__non_self<_ShowerProb>>>
    explicit shower_engine(_ShowerProb&& prob) : _M_trial{new obj<_ShowerProb>{std::forward<_ShowerProb>(prob)}}
    { _M_not_showered.reserve(64);}
    
    
    /*   shower evolution (at the moment it is just the LC+ approximation)   */
    void evolve(double tmax, shower_history<_Proc>& hist)
    {
      hist.k2T_cutoff = sqr(_M_opts.kT_cutoff);
      hist.max_color_suppression = _M_opts.max_color_suppression;
      _M_evolve_lcplus(hist, tmax);
    }

    /*   same like evolve but it returns the history   */
    shower_history<_Proc>& operator()(double tmax, shower_history<_Proc>& hist) {
      this->evolve(tmax, hist);
      return hist;
    }
    
    /*   Setting/getting engine options     */
    void options(const shower_engine_opts& opts) { _M_opts = opts;}
    const shower_engine_opts& options() const { return _M_opts;}
    
    /*   set the random number seed   */
    void seed(random_engine_type::result_type s) { _M_random.seed(s);}
    
  private:
    /*   random engine   */
    random_engine_type _M_random;
    
    /*   engine parameters    */
    shower_engine_opts _M_opts;
    
    /*   calculates color overlaps in the evolution    */
    color_overlap_su3 _M_overlap_su3;
    color_overlap_u3 _M_overlap_u3;
    
    /*   some workspace for keep tracking of the unshowered nodes   */
    std::vector<typename shower_history<_Proc>::iterator> _M_not_showered;
    
  private:
    /*    make one shower step   */
    typename shower_history<_Proc>::iterator
    _M_make_one_step(shower_history<_Proc>&, unsigned int, typename shower_history<_Proc>::iterator, double);
    
    /*   LC+ approximation of the shower evolution     */
    void _M_evolve_lcplus(shower_history<_Proc>&, double);
    
    
    /*   abstract class for type erasing   */
    struct base {
      virtual bool make_trial(shower_stage<_Proc>&, random_engine_type&, double, double) = 0;
      virtual __details::thrs_weight threshold(double, double) const = 0;
    };
    
    
    /*   this stores an instanse of the actual splitting probability   */
    template<class _ShowerProb>
    struct obj : base
    {
      explicit obj(_ShowerProb&& s) : prob{std::forward<_ShowerProb>(s)} {}
      
      bool make_trial(shower_stage<_Proc>& s, random_engine_type& r, double tmU, double tmL) {
        return prob.make_trial(s, r, tmU, tmL);
      }
      
      __details::thrs_weight threshold(double tmH, double tmL) const {
        return prob.threshold(tmH, tmL);
      }
      
      _ShowerProb prob;
    };
    
    
    /*   we put an instance of obj into an cloned_ptr   */
    cloned_ptr<base> _M_trial;
  };
  
  
  template<class _Proc> typename shower_history<_Proc>::iterator
  shower_engine<_Proc>::_M_make_one_step(shower_history<_Proc>& hist, unsigned int hcsidx,
                                         typename shower_history<_Proc>::iterator current,
                                         double tmax)
  {
    using iterator = typename shower_history<_Proc>::iterator;
    
    /*     make trial shower emissions    */
    bool is_possible = _M_trial -> make_trial(*current, _M_random, tmax, current->current_time);
    
    
    /*    calculating the threshold weight    */
    if(is_hadron_collision<_Proc>) {
      if(_M_opts.threshold.is_enabled)
        current->weights.apply_threshold_correction(_M_trial -> threshold(current->next_time, current->current_time));
    }

    /*   if there is no splitting we return with the current state    */
    if(!is_possible) return current;
    

    /*   selecting the color operator    */
    double weight = 0.0;
    
    if(current->is_color_u3) weight = __details::selecting_color_operator_u3(_M_random, *current);
    else weight = __details::selecting_color_operator_su3(_M_random, *current);
    
    /*    appending an uninitialized node to the history     */
    iterator next = hist.append(current);
  
    next->weights = current->weights;
    next->weights.unitary.main *= weight;
   
    next->is_color_u3 = current->is_color_u3;
    next->scale = current->vars.scale;
    next->current_time = current->next_time;
    next->next_time = tmax;
    next->vetos = current->vetos;
    
    /*     assigning the new momenta and flavors    */
    auto boost = __details::shower_mapping(next->state, current->state, current->vars);
    
    /*   reference momentum that helps to define shower time 
     *   we always boost it, that makes the whole thing lorentz invariant
     */
    next->qnull = boost(current->qnull);

    
    /*   change color group (SU(Nc) ==> U(Nc))    */
    if(next->is_color_u3 == false) {
      unsigned max_color_suppr = hcsidx + _M_opts.max_color_suppression;
      
      if(next->state.color_suppression_index >= max_color_suppr) {
        next->weights.unitary.main *= _M_overlap_su3(next->state);
        next->weights.unitary.main /= _M_overlap_u3(next->state);
        next->is_color_u3 = true;
      }
    }
    
    
    return next;
  }
  
 
  template<class _Proc>
  void shower_engine<_Proc>::_M_evolve_lcplus(shower_history<_Proc>& hist, double tmF)
  {
    using iterator = typename shower_history<_Proc>::iterator;

    /*    save all the leaf node of the tree   */
    _M_not_showered.resize(0);
    
    for(auto ip = hist.begin(); ip != hist.end(); ++ip)
      if(tmF > ip->next_time && std::abs(ip->weights.unitary.main) > 0.0)
      {
        if(ip->vetos.min_k2T < hist.k2T_cutoff)
          ip->vetos.min_k2T = hist.k2T_cutoff;
        
        _M_not_showered.push_back(ip);
      }
    
    /*   shower every events   */
    for(iterator& start : _M_not_showered)
    {
      unsigned int hcsidx = start.hard().state.color_suppression_index;
      
      /*    change color group of the starting stage, SU(Nc) ==> U(Nc), if it is needed   */
      if(start->is_color_u3 == false) {
        unsigned max_color_suppr = hcsidx + _M_opts.max_color_suppression;
        
        if(start->state.color_suppression_index >= max_color_suppr) {
          start->weights.unitary.main *= _M_overlap_su3(start->state);
          start->weights.unitary.main /= _M_overlap_u3(start->state);
          start->is_color_u3 = true;
        }
      
        /*    if the color weight is zero, stop showering        */
        if(std::abs(start->weights.unitary.main) == 0.0) continue;
      }
      
      /*    start the evolution   */
      iterator current, next = start;
      
      do {
        current = next;
        next = _M_make_one_step(hist, hcsidx, current, tmF);
      } while(current != next && std::abs(next->weights.unitary.main) > 0.0);
    }
  }
  

}   /*   namespace duct   */

#endif
