/*  Copyright © 2015 Zoltán Nagy. All rights reserved.
 *  Created by Zoltán Nagy on 28/10/15.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */
#ifndef __mc_hard_state_h__
#define __mc_hard_state_h__ 1


/*   std includes   */

/*   project headers   */
#include <bits/mc-shower-history.h>
#include <cxx/cloned-ptr.h>



namespace duct {
  
  namespace __details {
    
    /*    Calculating the hard threshold weights.
     *    The event has to be set already.
     */
    void hard_threshold_weight(shower_stage<dis>&, const model_ref&, const event<dis>&, double);
    void hard_threshold_weight(shower_stage<hhc>&, const model_ref&, const event<hhc>&, double);

    constexpr void hard_threshold_weight(shower_stage<epa>&, const model_ref&, const event<epa>&, double) {}
    constexpr void hard_threshold_weight(shower_stage<decay>&, const model_ref&, const event<decay>&, double) {}
  
  }  //  namespace __details
  
  
  
  /*   This object is for store (and use) generator of the hard states.
   *   The actual type of this generator is not visible it is erases,
   *   but one istance of the hard state generator is stored in this class.
   */
  template<class _Proc>
  class hard_state
  {
    template<typename T> static constexpr bool
    __non_self = (!std::is_same<hard_state, typename std::decay<T>::type>::value
                  && !std::is_base_of<hard_state, typename std::decay<T>::type>::value);

  public:
    /*   default constructor (no managed object)   */
    hard_state() = default;
    hard_state(hard_state&&) = default;
    hard_state(const hard_state&) = default;
    
    /*   crate from a hard state generator   */
    template<typename _Hard, typename _X = typename std::enable_if_t<__non_self<_Hard>>>
    explicit hard_state(_Hard&& h) : _M_hard{new obj<_Hard>{std::forward<_Hard>(h)}} {}
    
    
    /*    generate the hard state    */
    shower_history<_Proc>& operator()(shower_history<_Proc>&);
    /*   set the random number seed   */
    void seed(random_engine_type::result_type s) {
      _M_random.seed(s);
    }
    
  private:
    /*   abstract class for type erasing   */
    struct base {
      virtual void generate(random_engine_type&, shower_history<_Proc>&) = 0;
    };
    
    /*   this stores an instanse of the actual hard process   */
    template<class _Hard>
    struct obj : base
    {
      explicit obj(_Hard&& h)
      : hard{std::forward<_Hard>(h)} {}
 
      void generate(random_engine_type& r, shower_history<_Proc>& h) {
        hard.generate(r, h);
      }
      
      _Hard hard;
    };
    
    /*   we put an instance of obj<_Hard> into "any"   */
    cloned_ptr<base> _M_hard;
    
    /*   random engine   */
    random_engine_type _M_random;
  };
 
  
  
  template<class _Proc>
  shower_history<_Proc>& hard_state<_Proc>::operator()(shower_history<_Proc>& hist)
  {
    /*   make sure that we have a fresh start    */
    hist.new_history();
    
    /*   Generating the hard event.
     *   Our requirement to fill the following members of class shawer_state:
     *       ==> state (only: momentum, flavor, color, colorP, color_suppression_index)
     *       ==> weight.unitary.main
     *       ==> scale
     *       ==> vetos
     *
     *   The other members like current_time, next_time, thrs_weight,... are calculated
     *   in this function.
     */
    _M_hard -> generate(_M_random, hist);
    
    
    /*   completing the hard states    */
    for(auto& hard : hist.hard())
    {
      auto mdl = hard.state.get_model();

      /*   setting the dead colors  */
      for(auto& p : hard.state)
      {
        /*   assigning dead colors for non-QCD particles    */
        if(0 == mdl.color(p.flavor))
          p.color = p.colorP = particle_state::colorlink::dead();
      }
      
      /*   make the helth report   */
      hard.state.make_health_report();

      /*   We always start with SU(Nc) states.   */
      hard.is_color_u3 = false;
      
      /*   reference momentum to define shower time    */
      hard.qnull = hard.state[-1].momentum + hard.state[0].momentum;
           
      /*    shower time    */
      hard.current_time = hard.next_time = std::log(hard.qnull.mag2()/hard.scale);
      
      /*   assigning threshold weights    */
      __details::hard_threshold_weight(hard, mdl, hard.state, hard.scale);
    }
   
    
    return hist;
  }

  
  
  
}

#endif