/*  Copyright (C) 2012-2015 Zoltán Nagy
 *  Created by Zoltán Nagy on 22/2/12.
 *  
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>. 
 */
#ifndef __qcd_color_overlap_h__
#define __qcd_color_overlap_h__ 1


/*   std includes   */
#include <list>
#include <cmath>
#include <vector>


/*   project headers   */
#include <bits/qcd-color.h>
#include <bits/duct-event.h>



namespace duct {
  
  /*   forward declaration of color overlap classes   */
  class color_overlap_u3;
  class color_overlap_su3;

  
  namespace __details {
    
    class color_graph_su3
    {
    public:
      /*    we use the __partile_state color link type   */
      using colorlink = particle_state::colorlink;
      
      /*    class for representing a color state      */
      struct parton
      {
        /*  Multiplicity of the parton insertion:
         *    => multi == 0 inactive removed from the list
         *    => multi > 0 the multiplicity of the parton
         *       (we can have many parallel gluon)
         */
        unsigned int multi;
        
        /* These members for gluons. We use multi-gluons and these
         * are the coefficients of the corresponding Fierz identity.
         */
        double A, B;
        
        /*    color connections    */
        colorlink color, colorP;
      };
      
      /*    default constructor    */
      color_graph_su3() : _M_number_of_in{0u}, _M_active{0u}, _M_cf_pow{0}, _M_nc_pow{0}, _M_wgt{1.0} {}
      
      /*   non standard constructor, creating from event<...>    */
      template<std::size_t _Nin> color_graph_su3(const event_base<_Nin>& p) {
        auto qcd = p.template data<qcd_out>();
        this->assign(qcd.begin()-_Nin, qcd.begin(), qcd.end());
      }
      
      /*   assignement operator from event   */
      template<std::size_t _Nin>
      color_graph_su3& operator=(const event_base<_Nin>& p) {
        auto qcd = p.template data<qcd_out>();
        this->assign(qcd.begin()-_Nin, qcd.begin(), qcd.end());
        return *this;
      }
      
      /*   returns true if it is done and the overlap is calculated   */
      bool is_finished() const { return !_M_active;}
      
      /*   this applys all the simple rules   */
      void reduce();
      
    private:
      /*   assigment functions   */
      void assign(const particle_state *, const particle_state *, const particle_state *);
      
      /*   rules to simplify the graps   */
      void ruleG();
      void ruleQ();
      void ruleA();
      void ruleT();
      void ruleS();
      void ruleF(color_graph_su3&);
      
      /*   some helper   */
      parton * zero() { return _M_graph.data() + _M_number_of_in - 1;}
      parton * start() { return _M_graph.data();}
      parton * end() { return _M_graph.data() + _M_graph.size();}
      
    private:
      /*    storage for the partons   */
      std::vector<parton> _M_graph;
      
      /*   number of incoming colored objects and active partons   */
      unsigned int _M_number_of_in, _M_active;
      
      /*   powers of the Cf and Nc factors   */
      int _M_cf_pow, _M_nc_pow;
      
      /*   weight of the graph   */
      double _M_wgt;
      
      /*   friend declaration   */
      friend class duct::color_overlap_su3;
    };
    
    
    class color_graph_u3
    {
    public:
      /*    we use the __partile_state color link type   */
      using colorlink = particle_state::colorlink;
      
      /*    class for representing a color state      */
      struct parton
      {
        /*  Multiplicity of the parton insertion:
         *    => active == flase inactive removed from the list
         *    => active == true still in the list
         */
        bool active;
        
        /*    color connections    */
        colorlink color, colorP;
      };
      
      /*    default constructor    */
      color_graph_u3() : _M_number_of_in{0u}, _M_active{0u}, _M_nc_pow{0} {}
      
      /*   non standard constructor, creating from event<...>    */
      template<std::size_t _Nin> color_graph_u3(const event_base<_Nin>& p) {
        auto qcd = p.template data<qcd_out>();
        this->assign(qcd.begin()-_Nin, qcd.begin(), qcd.end());
      }
      
      /*   assignement operator from event   */
      template<std::size_t _Nin>
      color_graph_u3& operator=(const event_base<_Nin>& p) {
        auto qcd = p.template data<qcd_out>();
        this->assign(qcd.begin()-_Nin, qcd.begin(), qcd.end());
        return *this;
      }
      
      /*   returns true if it is done and the overlap is calculated   */
      bool is_finished() const { return !_M_active;}
      
      /*   this applys all the simple rules   */
      void reduce();
      
    private:
      /*   assigment functions   */
      void assign(const particle_state *, const particle_state *, const particle_state *);
      
      /*   remove a gluon    */
      void remove_gluon(int r)
      {
        parton *p = _M_graph.data() + _M_number_of_in -1;
        
        p[p[r].color.prev].color.next = p[r].color.next;
        p[p[r].color.next].color.prev = p[r].color.prev;
        
        p[p[r].colorP.prev].colorP.next = p[r].colorP.next;
        p[p[r].colorP.next].colorP.prev = p[r].colorP.prev;
        
        p[r].active = false;
        --_M_active;
      }
      
      /*   rules to simplify the graps   */
      void ruleG();
      void ruleQ();
      void ruleA();
      void ruleT();
      void ruleS();
      void ruleF();
      
      /*   some helper   */
      parton * zero() { return _M_graph.data() + _M_number_of_in - 1;}
      parton * start() { return _M_graph.data();}
      parton * end() { return _M_graph.data() + _M_graph.size();}
      
    private:
      /*    storage for the partons   */
      std::vector<parton> _M_graph;
      
      /*   number of incoming colored objects and active partons   */
      unsigned int _M_number_of_in, _M_active;
      
      /*   power of the Nc factor   */
      int _M_nc_pow;
      
      /*   friend declaration   */
      friend class duct::color_overlap_u3;
    };
    
  }   //   namespace __details
  
  
  
  /*    callable object class for calculating SU(3) color overlap   */
  class color_overlap_su3
  {
  public:
    /*    calculate the weight of a color configuration    */
    template<std::size_t _Nin> double operator()(const __details::event_base<_Nin>& p) {
      return this->calculate(__details::color_graph_su3(p));
    }

    /*   calculate the overlap of a given graph   */
    double calculate(__details::color_graph_su3&&);

  private:
    //----- we use list to store the graphs -----
    std::list<__details::color_graph_su3>::iterator _M_begin, _M_end;
    std::list<__details::color_graph_su3> _M_graphs;
  };

  
  /*    callable object class for calculatig U(3) color overlap     */
  class color_overlap_u3
  {
  public:
    /*    calculate the weight of a color configuration    */
    template<std::size_t _Nin> double operator()(const __details::event_base<_Nin>& p)
    {
      _M_graph = p;
 
      do {
        _M_graph.reduce();
        if(_M_graph._M_active == 0) break;
        _M_graph.ruleF();
      } while(true);
      
      return std::pow(qcd::Nc, _M_graph._M_nc_pow);
    }
    
  private:
    /*   we need only one graph as working place   */
    __details::color_graph_u3 _M_graph;
  };
  
}  //   namespace duct

#endif
