/*  Copyright (C) 2014 Zoltán Nagy
 *  Created by Zoltán Nagy on 2/6/14.
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*   std includes   */

/*   project headers  */
#include "bits/qcd-color.h"
#include "bits/mc-splitting-prob.h"
#include "cxx/math-support.h"


namespace duct {
  
  namespace __details {
    
    void counting_color_connections(stackvector<std::pair<int,double>, 4>& spectator, int emitter, const particle_state *p, const model_ref& mdl)
    {
      spectator.clear();
      
      switch(mdl.color(p[emitter].flavor))
      {
        case 8:
        {
          double cfac = 0.25*qcd::Ca;
         
          spectator.push_back({p[emitter].color.prev, cfac});
          spectator.push_back({p[emitter].color.next, cfac});
          
          if(p[emitter].colorP.prev == spectator[0].first) spectator[0].second += cfac;
          else if(p[emitter].colorP.prev == spectator[1].first) spectator[1].second += cfac;
          else spectator.push_back({p[emitter].colorP.prev, cfac});
          
          if(p[emitter].colorP.next == spectator[0].first) spectator[0].second += cfac;
          else if(p[emitter].colorP.next == spectator[1].first) spectator[1].second += cfac;
          else spectator.push_back({p[emitter].colorP.next, cfac});
        } break;
          
        case 3:
        {
          double cfac = 0.5*qcd::Cf;
          
          spectator.push_back({p[emitter].color.next, cfac});
          
          if(p[emitter].colorP.next == spectator[0].first) spectator[0].second += cfac;
          else spectator.push_back({p[emitter].colorP.next, cfac});
        } break;
        
        case -3:
        {
          double cfac = 0.5*qcd::Cf;
        
          spectator.push_back({p[emitter].color.prev, cfac});
          
          if(p[emitter].colorP.prev == spectator[0].first) spectator[0].second += cfac;
          else spectator.push_back({p[emitter].colorP.prev, cfac});
        } break;
      }
    }
    
    
    
    /*    Gauss integrals    */
    const gauss_integral<10u> splitting_prob_base<true>::threshold_integralT{};
    static const gauss_integral<10u> _threshold_integralZ{};
    
    
    
    /*   Functions for calculating the threshold weights for kT ordering.    */
    template<>
    thrs_weight splitting_prob_base<true>::
    threshold<ordering::kT>(const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      thrs_weight result{0.0,0.0};
      
      double q2 = as.q2(tau);
      double v = q2/pars->Q2;
      
      /*   the z limits are scheme dependent    */
      double zmin = 4.0/sqr(std::sqrt(v) + std::sqrt(4.0+v));
      double zmax = (1.0-pars->eta)*0.999 + pars->eta;
    
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2);
      
      
      /*   integrand of the main term    */
      auto color_diag = [=](double z) -> double {
        /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
         *   In this case the integrand is simple.
         */
        double oz = 1.0-z;
        
        if(z < pars->eta)
          return (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca)/oz;
        
        
        /*   Otherwise we have a more complicated function   */
        double coll, fchc, ratio, soft = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca);
        
        /*   soft term    */
        if(z >= 0.9999) {
          ratio = pars->aux->parton(fa, pars->eta/0.9999, q2)/pdf;
          soft *= (1.0 - ratio*z)/0.0001;
        }
        else {
          ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
          soft *= (1.0 - ratio*z)/oz;
        }
        
        
        /*   collinear terms : quark emitter   */
        if(fa) {
          coll  = qcd::Cf*oz*ratio;
          fchc  = qcd::Tr*(1.0-2.0*z*oz);
          fchc *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
        }
        
        /*   collinear terms : gluon emitter   */
        else {
          coll = 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
          fchc = qcd::Cf*(z + 2.0*oz/z);
          fchc *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
        }
        
        return soft - coll - fchc;
      };
      
      
      /*    color diagonal integral    */
      result.weight_main = _threshold_integralZ([=](double xi) -> double {
        double z = (1.0-zmax)*0.5*(1.0+xi) + zmax;
        return (1.0-zmax)*0.5*color_diag(z);
      });
      
      
      if(zmax > zmin)
      {
        double J = 0.5*std::log((1.0-zmin)/(1.0-zmax));
       
        result.weight_main += _threshold_integralZ([=](double xi) -> double {
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*J);
          return (1.0-z)*J*color_diag(z);
        });
      }
      
    
      /*    color off diagonal integral    */
      double J = 0.5*std::log(1.0/(1.0-zmin));
      
      result.weight_delta = _threshold_integralZ([=](double xi) -> double {
        /*   integral transformation    */
        double z = 1.0-1.0*std::exp(-(1.0+xi)*J);
        
        /*    sum over the specators  (only LC+ connections)    */
        double oz = 1.0-z, dsum = 0.0;
        
        for(const auto& spec : pars->spectators)
          if(spec.spectator > 0) {
            double xi = 1.0-spec.xi;
            dsum += 2.0*spec.cfac*(1.0/oz - xi/std::sqrt(sqr(oz*xi) + sqr(2.0*v*z/oz)*spec.xi));
          }
        
        /*   PDF factor    */
        if(z > pars->eta)
          dsum *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
        
        return oz*J*dsum;
      });
    
      
      return result;
    }
    
    
    
    /*   Functions for calculating the threshold weights for lambda ordering.    */
    template<>
    thrs_weight splitting_prob_base<true>::
    threshold<ordering::lambda>(const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      thrs_weight result{0.0,0.0};
      
      double q2 = as.q2(tau);
      double v = q2/pars->Q2;
      
      /*   the z limits are scheme dependent    */
      double zmin = 1.0/(1.0+v);
      double zmax = 1.0-pars->mT/v;
      double umin = std::max(0.0, 1.0-pars->mTmax/v);
      double umax = std::min(zmin, zmax);
      
      if(zmin < umin) zmin = umin;

      
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2);
      double asb = as.ask(tau)*qcd::beta1(as.nf(tau));
      
      
      /*    color diagonal integral    */
      if(zmax > zmin)
      {
        double J = 0.5*std::log((1.0-zmin)/(1.0-zmax));
        
        result.weight_main = _threshold_integralZ([=](double xi) -> double {
          /*   integral transformation    */
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*J);
          double oz = 1.0-z;
          
          /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
           *   In this case the integrand is simple.
           */
          if(z < pars->eta) {
            double soft = (fa ? 2.0*qcd::Cf : 2.0*qcd::Ca)/oz;
            return oz*J*soft/(1.0+asb*std::log(oz));
          }
          
          /*   Otherwise we have a more complicated function   */
          double ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
          double soft, coll, fchc;
          
          /*   soft term    */
          soft = (1.0 - ratio*z)*(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca)/oz;
          soft /= 1.0+asb*std::log(oz);
          
          /*   quark emitter   */
          if(fa) {
            coll  = qcd::Cf*oz*ratio;
            fchc  = qcd::Tr*(1.0-2.0*z*oz);
            fchc *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
          }
          
          /*   gluon emitter   */
          else {
            coll = 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
            fchc = qcd::Cf*(z + 2.0*oz/z);
            fchc *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
          }
          
          return oz*J*(soft - coll - fchc);
        });
      }
      
      /*    color off diagonal integral    */
      if(umax > umin)
      {
        double J = 0.5*std::log((1.0-umin)/(1.0-umax));
        
        result.weight_delta = _threshold_integralZ([=](double xi) -> double {
          /*   integral transformation    */
          double z = 1.0-(1.0-umin)*std::exp(-(1.0+xi)*J);
          
          /*    sum over the specators  (only LC+ connections)    */
          double oz = 1.0-z, dsum = 0.0;
          
          for(const auto& spec : pars->spectators)
            if(spec.spectator > 0) {
              double xi = 1.0-spec.xi;
              dsum += 2.0*spec.cfac*(1.0/oz - xi/std::sqrt(sqr(oz*xi) + sqr(2.0*v*z)*spec.xi));
            }
          
          /*   PDF factor   */
          if(z > pars->eta)
            dsum *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
          
          
          return oz*J*dsum/(1.0+asb*std::log(1.0-z));
        });
      }
      
      
      return result;
    }
    
 
    
    /*   Functions for calculating the threshold weights for angular ordering.    */
    template<>
    thrs_weight splitting_prob_base<true>::
    threshold<ordering::angular>(const hard_params<ini> *pars, double tau)
    {
      const qcd_coupling& as = pars->mdl.alpha_qcd();
      thrs_weight result{0.0,0.0};
      
      double q2 = as.q2(tau);
      double zmin = 0.0, zmax = 1.0, umin = 0.0, umax= 0.99999, v = q2/pars->Q2;
      
      
      /*   the z limits are scheme dependent    */
      zmin = std::min(1.0, 1.0/v);
      zmax = 1.0-std::sqrt(pars->mT/v);
      umin = std::max(0.0, 1.0-std::sqrt(pars->mTmax/v));
      umax = std::min(zmin, zmax);
      if(zmin < umin) zmin = umin;

      
      /*   defining the integrands of the z integrals    */
      int fa = -pars->fl;
      double pdf = pars->aux->parton(fa, pars->eta, q2);
      double asb = 2.0*as.ask(tau)*qcd::beta1(as.nf(tau));
      
      
      /*    color diagonal integral    */
      if(zmax > zmin)
      {
        double J = 0.5*std::log((1.0-zmin)/(1.0-zmax));
        
        result.weight_main = _threshold_integralZ([=](double xi) -> double {
          /*   integral transformation    */
          double z = 1.0-(1.0-zmin)*std::exp(-(1.0+xi)*J);
          double oz = 1.0-z;
          
          /*   When z < eta then f(eta/z, q2) = 0 for every flavors.
           *   In this case the integrand is simple.
           */
          if(z < pars->eta)
            return J*(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca)/(1.0+asb*std::log(oz));
    
          
          /*   Otherwise we have a more complicated function   */
          double ratio = pars->aux->parton(fa, pars->eta/z, q2)/pdf;
          double soft, coll, fchc;
          
          /*   soft term    */
          soft = (1.0 - ratio*z)*(fa ? 2.0*qcd::Cf : 2.0*qcd::Ca)/oz;
          soft /= 1.0+asb*std::log(oz);
          
          /*   quark emitter   */
          if(fa) {
            coll  = qcd::Cf*oz*ratio;
            fchc  = qcd::Tr*(1.0-2.0*z*oz);
            fchc *= pars->aux->parton(0, pars->eta/z, q2)/pdf;
          }
          
          /*   gluon emitter   */
          else {
            coll = 2.0*qcd::Ca*(oz/z+z*oz)*ratio;
            fchc = qcd::Cf*(z + 2.0*oz/z);
            fchc *= pars->aux->parton(81, pars->eta/z, q2)/pdf;
          }
          
          return oz*J*(soft - coll - fchc);
        });
      }
      
      /*    color off diagonal integral    */
      if(umax > umin)
      {
        double J = 0.5*std::log((1.0-umin)/(1.0-umax));
        
        result.weight_delta = _threshold_integralZ([=](double xi) -> double {
          /*   integral transformation    */
          double dsum = 0.0, z = 1.0-(1.0-umin)*std::exp(-(1.0+xi)*J);
          
          /*    sum over the specators  (only LC+ connections)    */
          for(const auto& spec : pars->spectators)
            if(spec.spectator > 0) {
              double xi = 1.0-spec.xi;
              dsum += 2.0*spec.cfac*(1.0 - xi/std::sqrt(sqr(xi) + sqr(2.0*v*z)*spec.xi));
            }
          
          /*   PDF factor  & alpha_s factors    */
          if(z > pars->eta)
            dsum *= 1.0 - pars->aux->parton(fa, pars->eta/z, q2)*z/pdf;
          
          return J*dsum/(1.0+asb*std::log(1.0-z));
        });
      }
      
      
      
      return result;
    }
    
    
  }   /*   namespace __details   */


}

